#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_region.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Module used to select cells based on many different criteria.

This module implements the class EclRegion which can be used to select
cells in a grid matching a criteria. A wide range of different
criteria are supported. Many of the special functions for implementing
mathematical operations are implemented, so that regions can be
combined e.g. with logical &.

When the selection process is complete the region instance can be
queried for the corresponding list of indices.
"""
import ctypes

from cwrap import BaseCClass

import ecl
from ecl.util.util import monkey_the_camel
from ecl.util.util import IntVector

from ecl import EclPrototype
from ecl.grid.faults import Layer
from ecl import EclDataType
from ecl.eclfile import EclKW
from ecl.util.geometry import CPolyline


def select_method(select):
    """
    The select_method method decorator is applied to all the
    select_xxx() methods. The purpose of this decorator is to
    allow the select_xxx() methods to have an optional argument
    @intersect. If the @intersect argument is True the results of
    the current select method will be and'ed with the current
    selection, instead of or'ed which is the default.

    Consider this example:

       region = EclRegion( grid , False )
       region.select_islice(0 , 5)     # Selects all cells with i:[0:5]
       region.select_jslice(0 , 5)     # Selects all cells with j:[0:5]

    When these two calls have completed selection will contain all
    the cells which are either in i-interval [0:5] or in
    j-interval [0:5]. If we supply the @intersect argument in the
    second call the j selection will only be applied to the cells
    in i:[0:5] interval:

       region = EclRegion( grid , False )
       region.select_islice(0 , 5)     # Selects all cells with i:[0:5]
       region.select_jslice(0 , 5)     # Selects all cells with j:[0:5] AND i:[0:5]
    """

    def select_wrapper(self , *args ,  **kwargs):
        intersect = 'intersect' in kwargs and kwargs['intersect']
        if intersect:
            new_region = EclRegion( self.grid , False )
            select(new_region , *args )

            self &= new_region
        else:
            select(self , *args )

    return select_wrapper

class EclRegion(BaseCClass):
    TYPE_NAME = "ecl_region"
    _alloc                      = EclPrototype("void* ecl_region_alloc( ecl_grid , bool )", bind = False)
    _alloc_copy                 = EclPrototype("ecl_region_obj ecl_region_alloc_copy( ecl_region )")

    _set_kw_int                 = EclPrototype("void ecl_region_set_kw_int( ecl_region , ecl_kw , int, bool) ")
    _set_kw_float               = EclPrototype("void ecl_region_set_kw_float( ecl_region , ecl_kw , float, bool ) ")
    _set_kw_double              = EclPrototype("void ecl_region_set_kw_double( ecl_region , ecl_kw , double , bool) ")
    _shift_kw_int               = EclPrototype("void ecl_region_shift_kw_int( ecl_region , ecl_kw , int, bool) ")
    _shift_kw_float             = EclPrototype("void ecl_region_shift_kw_float( ecl_region , ecl_kw , float, bool ) ")
    _shift_kw_double            = EclPrototype("void ecl_region_shift_kw_double( ecl_region , ecl_kw , double , bool) ")
    _scale_kw_int               = EclPrototype("void ecl_region_scale_kw_int( ecl_region , ecl_kw , int, bool) ")
    _scale_kw_float             = EclPrototype("void ecl_region_scale_kw_float( ecl_region , ecl_kw , float, bool ) ")
    _scale_kw_double            = EclPrototype("void ecl_region_scale_kw_double( ecl_region , ecl_kw , double , bool) ")
    _sum_kw_int                 = EclPrototype("int ecl_region_sum_kw_int( ecl_region , ecl_kw , bool) ")
    _sum_kw_float               = EclPrototype("float ecl_region_sum_kw_float( ecl_region , ecl_kw , bool ) ")
    _sum_kw_double              = EclPrototype("double ecl_region_sum_kw_double( ecl_region , ecl_kw , bool) ")
    _sum_kw_bool                = EclPrototype("int ecl_region_sum_kw_int( ecl_region , ecl_kw , bool) ")

    _free                       = EclPrototype("void ecl_region_free( ecl_region )")
    _reset                      = EclPrototype("void ecl_region_reset( ecl_region )")
    _select_all                 = EclPrototype("void ecl_region_select_all( ecl_region )")
    _deselect_all               = EclPrototype("void ecl_region_deselect_all( ecl_region )")
    _select_equal               = EclPrototype("void ecl_region_select_equal( ecl_region , ecl_kw , int )")
    _deselect_equal             = EclPrototype("void ecl_region_deselect_equal( ecl_region , ecl_kw , int)")
    _select_less                = EclPrototype("void ecl_region_select_smaller( ecl_region , ecl_kw , float )")
    _deselect_less              = EclPrototype("void ecl_region_deselect_smaller( ecl_region , ecl_kw , float )")
    _select_more                = EclPrototype("void ecl_region_select_larger( ecl_region , ecl_kw , float )")
    _deselect_more              = EclPrototype("void ecl_region_deselect_larger( ecl_region , ecl_kw , float )")
    _select_in_interval         = EclPrototype("void ecl_region_select_in_interval( ecl_region, ecl_kw , float , float )")
    _deselect_in_interval       = EclPrototype("void ecl_region_deselect_in_interval( ecl_region, ecl_kw, float , float )")
    _invert_selection           = EclPrototype("void ecl_region_invert_selection( ecl_region )")

    _select_box                 = EclPrototype("void ecl_region_select_from_ijkbox(ecl_region , int , int , int , int , int , int)")
    _deselect_box               = EclPrototype("void ecl_region_deselect_from_ijkbox(ecl_region , int , int , int , int , int , int)")
    _imul_kw                    = EclPrototype("void  ecl_region_kw_imul( ecl_region , ecl_kw , ecl_kw , bool)")
    _idiv_kw                    = EclPrototype("void  ecl_region_kw_idiv( ecl_region , ecl_kw , ecl_kw , bool)")
    _iadd_kw                    = EclPrototype("void  ecl_region_kw_iadd( ecl_region , ecl_kw , ecl_kw , bool)")
    _isub_kw                    = EclPrototype("void  ecl_region_kw_isub( ecl_region , ecl_kw , ecl_kw , bool)")
    _copy_kw                    = EclPrototype("void  ecl_region_kw_copy( ecl_region , ecl_kw , ecl_kw , bool)")
    _intersect                  = EclPrototype("void ecl_region_intersection( ecl_region , ecl_region )")
    _combine                    = EclPrototype("void ecl_region_union( ecl_region , ecl_region )")
    _subtract                   = EclPrototype("void ecl_region_subtract( ecl_region , ecl_region )")
    _xor                        = EclPrototype("void ecl_region_xor( ecl_region , ecl_region )")
    _get_kw_index_list          = EclPrototype("int_vector_ref ecl_region_get_kw_index_list( ecl_region , ecl_kw , bool )")
    _get_active_list            = EclPrototype("int_vector_ref ecl_region_get_active_list( ecl_region )")
    _get_global_list            = EclPrototype("int_vector_ref ecl_region_get_global_list( ecl_region )")
    _get_active_global          = EclPrototype("int_vector_ref ecl_region_get_global_active_list( ecl_region )")
    _select_cmp_less            = EclPrototype("void ecl_region_cmp_select_less( ecl_region , ecl_kw , ecl_kw)")
    _select_cmp_more            = EclPrototype("void ecl_region_cmp_select_more( ecl_region , ecl_kw , ecl_kw)")
    _deselect_cmp_less          = EclPrototype("void ecl_region_cmp_deselect_less( ecl_region , ecl_kw , ecl_kw)")
    _deselect_cmp_more          = EclPrototype("void ecl_region_cmp_deselect_more( ecl_region , ecl_kw , ecl_kw)")
    _select_islice              = EclPrototype("void ecl_region_select_i1i2( ecl_region , int , int )")
    _deselect_islice            = EclPrototype("void ecl_region_deselect_i1i2( ecl_region , int , int )")
    _select_jslice              = EclPrototype("void ecl_region_select_j1j2( ecl_region , int , int )")
    _deselect_jslice            = EclPrototype("void ecl_region_deselect_j1j2( ecl_region , int , int )")
    _select_kslice              = EclPrototype("void ecl_region_select_k1k2( ecl_region , int , int )")
    _deselect_kslice            = EclPrototype("void ecl_region_deselect_k1k2( ecl_region , int , int )")
    _select_deep_cells          = EclPrototype("void ecl_region_select_deep_cells( ecl_region , double )")
    _deselect_deep_cells        = EclPrototype("void ecl_region_select_deep_cells( ecl_region , double )")
    _select_shallow_cells       = EclPrototype("void ecl_region_select_shallow_cells( ecl_region , double )")
    _deselect_shallow_cells     = EclPrototype("void ecl_region_select_shallow_cells( ecl_region , double )")
    _select_small               = EclPrototype("void ecl_region_select_small_cells( ecl_region , double )")
    _deselect_small             = EclPrototype("void ecl_region_deselect_small_cells( ecl_region , double )")
    _select_large               = EclPrototype("void ecl_region_select_large_cells( ecl_region , double )")
    _deselect_large             = EclPrototype("void ecl_region_deselect_large_cells( ecl_region , double )")
    _select_thin                = EclPrototype("void ecl_region_select_thin_cells( ecl_region , double )")
    _deselect_thin              = EclPrototype("void ecl_region_deselect_thin_cells( ecl_region , double )")
    _select_thick               = EclPrototype("void ecl_region_select_thick_cells( ecl_region , double )")
    _deselect_thick             = EclPrototype("void ecl_region_deselect_thick_cells( ecl_region , double )")
    _select_active              = EclPrototype("void ecl_region_select_active_cells( ecl_region )")
    _select_inactive            = EclPrototype("void ecl_region_select_inactive_cells( ecl_region )")
    _deselect_active            = EclPrototype("void ecl_region_deselect_active_cells( ecl_region )")
    _deselect_inactive          = EclPrototype("void ecl_region_deselect_inactive_cells( ecl_region )")
    _select_above_plane         = EclPrototype("void ecl_region_select_above_plane( ecl_region  , double* , double* )")
    _select_below_plane         = EclPrototype("void ecl_region_select_below_plane( ecl_region  , double* , double* )")
    _deselect_above_plane       = EclPrototype("void ecl_region_deselect_above_plane( ecl_region, double* , double* )")
    _deselect_below_plane       = EclPrototype("void ecl_region_deselect_below_plane( ecl_region, double* , double* )")
    _select_inside_polygon      = EclPrototype("void ecl_region_select_inside_polygon( ecl_region , geo_polygon)")
    _select_outside_polygon     = EclPrototype("void ecl_region_select_outside_polygon( ecl_region , geo_polygon)")
    _deselect_inside_polygon    = EclPrototype("void ecl_region_deselect_inside_polygon( ecl_region , geo_polygon)")
    _deselect_outside_polygon   = EclPrototype("void ecl_region_deselect_outside_polygon( ecl_region , geo_polygon)")
    _set_name                   = EclPrototype("void ecl_region_set_name( ecl_region , char*)")
    _get_name                   = EclPrototype("char* ecl_region_get_name( ecl_region )")
    _contains_ijk               = EclPrototype("void ecl_region_contains_ijk( ecl_region , int , int , int)")
    _contains_global            = EclPrototype("void ecl_region_contains_global( ecl_region, int )")
    _contains_active            = EclPrototype("void ecl_region_contains_active( ecl_region , int )")
    _equal                      = EclPrototype("bool ecl_region_equal( ecl_region , ecl_region )")
    _select_true                = EclPrototype("void ecl_region_select_true( ecl_region , ecl_kw)")
    _select_false               = EclPrototype("void ecl_region_select_false( ecl_region , ecl_kw)")
    _deselect_true              = EclPrototype("void ecl_region_deselect_true( ecl_region , ecl_kw)")
    _deselect_false             = EclPrototype("void ecl_region_deselect_false( ecl_region , ecl_kw)")
    _select_from_layer          = EclPrototype("void ecl_region_select_from_layer( ecl_region , layer , int , int)")
    _deselect_from_layer        = EclPrototype("void ecl_region_deselect_from_layer( ecl_region , layer , int , int)")


    def __init__(self , grid , preselect):
        """
        Create a new region selector for cells in @grid.

        Will create a new region selector to select and deselect the
        cells in the grid given by @grid. The input argument @grid
        should be a EclGrid instance. You can start with either all
        cells, or no cells, selected, depending on the value of
        @preselect.
        """

        self.grid = grid
        self.active_index = False
        c_ptr = self._alloc( grid , preselect )
        super(EclRegion , self).__init__( c_ptr )


    def free(self):
        self._free( )


    def __eq__(self , other):
        return self._equal(other)

    def __hash__(self):
        return hash(hash(self.grid) + hash(self.active_index))


    def __deep_copy__(self , memo):
        """
        Creates a deep copy of the current region.
        """
        return self._alloc_copy( )

    def __nonzero__(self):
        global_list = self.get_global_list()
        return len(global_list) > 0

    def __bool__(self):
        return self.__nonzero__()

    def __iand__(self , other):
        """
        Will perform set intersection operation inplace.

        Will update the current region selection so that the elements
        selected in self are also selected in @other. Bound to the
        inplace & operator, i.e.

           reg1 &= reg2

        will eventually call this method.
        """
        if isinstance(other , EclRegion):
            self._intersect(  other)
        else:
            raise TypeError("Ecl region can only intersect with other EclRegion instances")

        return self


    def __isub__(self , other):
        """
        Inplace "subtract" one selection from another.

        Bound to reg -= reg2
        """
        if isinstance( other , EclRegion ):
            self._subtract(   other)
        else:
            raise TypeError("Ecl region can only subtract with other EclRegion instances")

        return self


    def __ior__(self , other):
        """
        Will perform set operation union in place.

        The current region selection will be updated to contain all
        the elements which are selected either in the current region,
        or in @other; bound to to inplace | operator, so you can write e.g.

            reg1 |= reg2

        to update reg1 with the selections from reg2.
        """
        if isinstance( other , EclRegion):
            self._combine(  other)
        else:
            raise TypeError("Ecl region can only be combined with other EclRegion instances")

        return self

    def __iadd__(self , other):
        """
        Combines to regions - see __ior__().
        """
        return self.__ior__( other )

    def __or__(self , other):
        """
        Creates a new region which is the union of @self and other.

        The method will create a new region which selection status is
        given by the logical or of regions @self and @other; the two
        initial regions will not be modified. Bound to the unary |
        operator:

            new_reg = reg1 | reg2

        """
        new_region = self.copy()
        new_region.__ior__( other )
        return new_region

    def __and__(self , other):
        """
        Creates a new region which is the intersection of @self and other.

        The method will create a new region which selection status is
        given by the logical and of regions @self and @other; the two
        initial regions will not be modified. Bound to the unary &
        operator:

            new_reg = reg1 & reg2
        """
        new_region = self.copy()
        new_region.__iand__( other )
        return new_region

    def __add__(self , other):
        """
        Unary add operator for two regions - implemented by __or__().
        """
        return self.__or__( other )


    def __sub__( self, other):
        """
        Unary del operator for two regions.
        """
        new_region = self.copy()
        new_region.__isub__( other )
        return new_region


    def union_with( self, other):
        """
        Will update self with the union of @self and @other.

        See doscumentation of __ior__().
        """
        return self.__ior__( other )

    def intersect_with( self, other):
        """
        Will update self with the intersection of @self and @other.

        See doscumentation of __iand__().
        """
        return self.__iand__( other )


    def copy( self ):
        return self.__deep_copy__( {} )

    def reset(self):
        """
        Clear selections according to constructor argument @preselect.

        Will clear all selections, depending on the value of the
        constructor argument @preselect. If @preselect is true
        everything will be selected after calling reset(), otherwise
        no cells will be selected after calling reset().
        """
        self._reset(  )

    ##################################################################


    @select_method
    def select_more( self , ecl_kw , limit , intersect = False):
        """
        Select all cells where keyword @ecl_kw is above @limit.

        This method is used to select all the cells where an arbitrary
        field, contained in @ecl_kw, is above a limiting value
        @limit. The EclKW instance must have either nactive or
        nx*ny*nz elements; if this is not satisfied method will fail
        hard. The datatype of @ecl_kw must be numeric,
        i.e. ECL_INT_TYPE, ECL_DOUBLE_TYPE or ECL_FLOAT_TYPE. In the
        example below we select all the cells with water saturation
        above 0.85:

           restart_file = ecl.EclFile( "ECLIPSE.X0067" )
           swat_kw = restart_file["SWAT"][0]
           grid = ecl.EclGrid( "ECLIPSE.EGRID" )
           region = ecl.EclRegion( grid , False )

           region.select_more( swat_kw , 0.85 )

        """
        self._select_more( ecl_kw , limit )



    def deselect_more( self , ecl_kw , limit):
        """
        Deselects cells with value above limit.

        See select_more() for further documentation.
        """
        self._deselect_more( ecl_kw , limit )

    @select_method
    def select_less( self , ecl_kw , limit , intersect = False):
        """
        Select all cells where keyword @ecl_kw is below @limit.

        See select_more() for further documentation.
        """
        self._select_less( ecl_kw , limit )

    def deselect_less( self , ecl_kw , limit):
        """
        Deselect all cells where keyword @ecl_kw is below @limit.

        See select_more() for further documentation.
        """
        self._deselect_less( ecl_kw , limit )

    @select_method
    def select_equal( self , ecl_kw , value , intersect = False):
        """
        Select all cells where @ecl_kw is equal to @value.

        The EclKW instance @ecl_kw must be of size nactive or
        nx*ny*nz, and it must be of integer type; testing for equality
        is not supported for floating point numbers. In the example
        below we select all the cells in PVT regions 2 and 4:

           init_file = ecl.EclFile( "ECLIPSE.INIT" )
           pvtnum_kw = init_file.iget_named_kw( "PVTNUM" , 0 )
           grid = ecl.EclGrid( "ECLIPSE.GRID" )
           region = ecl.EclRegion( grid , False )

           region.select_equal( pvtnum_kw , 2 )
           region.select_equal( pvtnum_kw , 4 )

        """
        if not ecl_kw.data_type.is_int():
            raise ValueError("The select_equal method must have an integer valued keyword - got:%s" % ecl_kw.typeName( ))
        self._select_equal( ecl_kw , value )


    def deselect_equal( self , ecl_kw , value ):
        """
        Select all cells where @ecl_kw is equal to @value.

        See select_equal() for further documentation.
        """
        if not ecl_kw.data_type.is_int():
            raise ValueError("The select_equal method must have an integer valued keyword - got:%s" % ecl_kw.typeName( ))
        self._deselect_equal( ecl_kw , value )

    @select_method
    def select_in_range( self , ecl_kw , lower_limit , upper_limit , select = False):
        """
        Select all cells where @ecl_kw is in the half-open interval [ , ).

        Will select all the cells where EclKW instance @ecl_kw has
        value in the half-open interval [@lower_limit ,
        @upper_limit). The input argument @ecl_kw must have size
        nactive or nx*ny*nz, and it must be of type ECL_FLOAT_TYPE.

        The following example will select all cells with porosity in
        the range [0.15,0.20):

           init_file = ecl.EclFile( "ECLIPSE.INIT" )
           poro_kw = init_file.iget_named_kw( "PORO" , 0 )
           grid = ecl.EclGrid( "ECLIPSE.GRID" )
           region = ecl.EclRegion( grid , False )

           region.select_in_range( poro_kw , 0.15, 0.20 )

        """
        self._select_in_interval( ecl_kw , lower_limit , upper_limit)

    def deselect_in_range( self , ecl_kw , lower_limit , upper_limit):
        """
        Deselect all cells where @ecl_kw is in the half-open interval [ , ).

        See select_in_range() for further documentation.
        """
        self._deselect_in_interval( ecl_kw , lower_limit , upper_limit)


    @select_method
    def select_cmp_less( self , kw1 , kw2 , intersect = False):
        """
        Will select all cells where kw2 < kw1.

        Will compare the ECLIPSE keywords @kw1 and @kw2, and select
        all the cells where the numerical value of @kw1 is less than
        the numerical value of @kw2. The ECLIPSE keywords @kw1 and
        @kw2 must both be of the same size, nactive or nx*ny*nz. In
        addition they must both be of type type ECL_FLOAT_TYPE. In the
        example below we select all the cells where the pressure has
        dropped:

           restart_file = ecl.EclFile("ECLIPSE.UNRST")
           pressure1 = restart_file.iget_named_kw( "PRESSURE" , 0)
           pressure2 = restart_file.iget_named_kw( "PRESSURE" , 100)

           region.select_cmp_less( pressure2 , pressure1)

        """
        self._select_cmp_less( kw1 , kw2 )

    def deselect_cmp_less( self , kw1 , kw2):
        """
        Will deselect all cells where kw2 < kw1.

        See select_cmp_less() for further documentation.
        """
        self._deselect_cmp_less( kw1 , kw2 )

    @select_method
    def select_cmp_more( self , kw1 , kw2 , intersect = False):
        """
        Will select all cells where kw2 > kw1.

        See select_cmp_less() for further documentation.
        """
        self._select_cmp_more( kw1 , kw2 )

    def deselect_cmp_more( self , kw1 , kw2):
        """
        Will deselect all cells where kw2 > kw1.

        See select_cmp_less() for further documentation.
        """
        self._deselect_cmp_more( kw1 , kw2 )

    @select_method
    def select_active( self , intersect = False):
        """
        Will select all the active grid cells.
        """
        self._select_active(  )

    def deselect_active( self ):
        """
        Will deselect all the active grid cells.
        """
        self._deselect_active(  )

    @select_method
    def select_inactive( self , intersect = False):
        """
        Will select all the inactive grid cells.
        """
        self._select_inactive(  )


    def deselect_inactive( self ):
        """
        Will deselect all the inactive grid cells.
        """
        self._deselect_inactive( )


    def select_all( self ):
        """
        Will select all the cells.
        """
        self._select_all( )

    def deselect_all( self ):
        """
        Will deselect all the cells.
        """
        self._deselect_all( )

    def clear( self ):
        """
        Will deselect all cells.
        """
        self.deselect_all()

    @select_method
    def select_deep( self , depth , intersect = False):
        """
        Will select all cells below @depth.
        """
        self._select_deep_cells(depth)

    def deselect_deep( self, depth):
        """
        Will deselect all cells below @depth.
        """
        self._deselect_deep_cells(depth)

    @select_method
    def select_shallow( self, depth , intersect = False):
        """
        Will select all cells above @depth.
        """
        self._select_shallow_cells(depth)

    def deselect_shallow( self, depth):
        """
        Will deselect all cells above @depth.
        """
        self._deselect_shallow_cells(depth)

    @select_method
    def select_small( self , size_limit , intersect = False):
        """
        Will select all cells smaller than @size_limit.
        """
        self._select_small( size_limit )

    def deselect_small( self , size_limit ):
        """
        Will deselect all cells smaller than @size_limit.
        """
        self._deselect_small( size_limit )

    @select_method
    def select_large( self , size_limit , intersect = False):
        """
        Will select all cells larger than @size_limit.
        """
        self._select_large( size_limit )

    def deselect_large( self , size_limit ):
        """
        Will deselect all cells larger than @size_limit.
        """
        self._deselect_large( size_limit )

    @select_method
    def select_thin( self , size_limit , intersect = False):
        """
        Will select all cells thinner than @size_limit.
        """
        self._select_thin( size_limit )

    def deselect_thin( self , size_limit ):
        """
        Will deselect all cells thinner than @size_limit.
        """
        self._deselect_thin( size_limit )

    @select_method
    def select_thick( self , size_limit , intersect = False):
        """
        Will select all cells thicker than @size_limit.
        """
        self._select_thick( size_limit )

    def deselect_thick( self , size_limit ):
        """
        Will deselect all cells thicker than @size_limit.
        """
        self._deselect_thick( size_limit )

    @select_method
    def select_box( self , ijk1 , ijk2 , intersect = False):
        """
        Will select all cells in box.

        Will select all the the cells in the box given by @ijk1 and
        @ijk2. The two arguments @ijk1 and @ijk2 are tuples (1,j,k)
        representing two arbitrary - diagonally opposed corners - of a
        box. All the elements in @ijk1 and @ijk2 are inclusive, i.e.

           select_box( (10,12,8) , (8 , 16,4) )

        will select the box defined by [8,10] x [12,16] x [4,8].
        """
        self._select_box( ijk1[0] , ijk2[0] , ijk1[1] , ijk2[1] , ijk1[2] , ijk2[2])

    def deselect_box( self , ijk1 , ijk2 ):
        """
        Will deselect all elements in box.

        See select_box() for further documentation.
        """
        self._deselect_box( ijk1[0] , ijk2[0] , ijk1[1] , ijk2[1] , ijk1[2] , ijk2[2])

    @select_method
    def select_islice( self , i1 , i2, intersect = False):
        """
        Will select all cells with i in [@i1, @i2]. @i1 and @i2 are zero offset.
        """
        self._select_islice( i1,i2)

    def deselect_islice( self , i1 , i2):
        """
        Will deselect all cells with i in [@i1, @i2]. @i1 and @i2 are zero offset.
        """
        self._deselect_islice( i1,i2)

    @select_method
    def select_jslice( self , j1 , j2 , intersect = False):
        """
        Will select all cells with j in [@j1, @j2]. @i1 and @i2 are zero offset.
        """
        self._select_jslice( j1,j2)

    def deselect_jslice( self , j1 , j2):
        """
        Will deselect all cells with j in [@j1, @j2]. @i1 and @i2 are zero offset.
        """
        self._deselect_jslice( j1,j2)

    @select_method
    def select_kslice( self , k1 , k2 , intersect = False):
        """
        Will select all cells with k in [@k1, @k2]. @i1 and @i2 are zero offset.
        """
        self._select_kslice( k1,k2)

    def deselect_kslice( self , k1 , k2):
        """
        Will deselect all cells with k in [@k1, @k2]. @i1 and @i2 are zero offset.
        """
        self._deselect_kslice( k1,k2)

    def invert( self ):
        """
        Will invert the current selection.
        """
        self._invert_selection(  )

    def __init_plane_select( self , n , p ):
        n_vec = ctypes.cast( (ctypes.c_double * 3)() , ctypes.POINTER( ctypes.c_double ))
        p_vec = ctypes.cast( (ctypes.c_double * 3)() , ctypes.POINTER( ctypes.c_double ))
        for i in range(3):
            n_vec[i] = n[i]
            p_vec[i] = p[i]
        return ( n_vec , p_vec )

    @select_method
    def select_above_plane( self , n , p , intersect = False):
        """
        Will select all the cells 'above' the plane defined by n & p.

        @n is the surface normal vector of the plane in question and
        @p is a point on the plane surface. The point @p should be
        given in (utm_x , utm_y , tvd) coordinates. The term 'above'
        means that the cell center has a positive distance to the
        plain; correspondingly 'below' means that the cell center has
        a negative disatnce to the plane.
        """
        (n_vec , p_vec) = self.__init_plane_select( n , p )
        self._select_above_plane( n_vec , p_vec )

    @select_method
    def select_below_plane( self , n , p , interscet = False):
        """
        Will select all the cells 'below' the plane defined by n & p.

        See method 'select_above_plane' for further documentation.
        """
        (n_vec , p_vec) = self.__init_plane_select( n , p )
        self._select_below_plane( n_vec , p_vec )

    def deselect_above_plane( self , n , p):
        """
        Will deselect all the cells 'above' the plane defined by n & p.

        See method 'select_above_plane' for further documentation.
        """
        (n_vec , p_vec) = self.__init_plane_select( n , p )
        self._deselect_above_plane( n_vec , p_vec )

    def deselect_below_plane( self , n , p):
        """
        Will deselect all the cells 'below' the plane defined by n & p.

        See method 'select_above_plane' for further documentation.
        """
        (n_vec , p_vec) = self.__init_plane_select( n , p )
        self._deselect_below_plane( n_vec , p_vec )

    @select_method
    def select_inside_polygon( self , points , intersect = False):
        """
        Will select all points inside polygon.

        Will select all points inside polygon specified by input
        variable @points. Points should be a list of two-element
        tuples (x,y). So to select all the points within the rectangle
        bounded by the lower left rectangle (0,0) and upper right
        (100,100) the @points list should be:

           points = [(0,0) , (0,100) , (100,100) ,  (100,0)]

        The elements in the points list should be (utm_x, utm_y)
        values. These values will be compared with the centerpoints of
        the cells in the grid. The selection is based the top k=0
        layer, and then extending this selection to all k values; this
        implies that the selection polygon will effectively be
        translated if the pillars are not vertical.
        """
        self._select_inside_polygon( CPolyline( init_points = points ))

    @select_method
    def select_outside_polygon( self , points , intersect = False):
        """
        Will select all points outside polygon.

        See select_inside_polygon for more docuemntation.
        """
        self._select_outside_polygon( CPolyline( init_points = points ))

    def deselect_inside_polygon( self , points ):
        """
        Will select all points outside polygon.

        See select_inside_polygon for more docuemntation.
        """
        self._deselect_inside_polygon( CPolyline( init_points = points ))

    def deselect_outside_polygon( self , points ):
        """
        Will select all points outside polygon.

        See select_inside_polygon for more docuemntation.
        """
        self._deselect_outside_polygon( CPolyline( init_points = points ))


    @select_method
    def select_true( self , ecl_kw , intersect = False):
        """
        Assume that input ecl_kw is a boolean mask.
        """
        self._select_true( ecl_kw )


    @select_method
    def select_false( self , ecl_kw , intersect = False):
        """
        Assume that input ecl_kw is a boolean mask.
        """
        self._select_false( ecl_kw )


    @select_method
    def select_from_layer(self , layer , k , value, intersect = False):
        """Will select all the cells in in @layer with value @value - at
        vertical coordinate @k.

        The input @layer should be of type Layer - from the
        ecl.ecl.faults.layer module. The k value must in the range
        [0,grid.nz) and the dimensions of the layer must correspond
        exactly to nx,ny of the grid.
        """
        grid = self.grid
        if k < 0 or k >= grid.getNZ():
            raise ValueError("Invalid k value:%d - must be in range [0,%d)" % (k , grid.getNZ()))

        if grid.getNX() != layer.getNX():
            raise ValueError("NX dimension mismatch. Grid:%d  layer:%d" % (grid.getNX() , layer.getNX()))

        if grid.getNY() != layer.getNY():
            raise ValueError("NY dimension mismatch. Grid:%d  layer:%d" % (grid.getNY() , layer.getNY()))

        self._select_from_layer( layer , k , value )



    #################################################################

    def scalar_apply_kw( self , target_kw , scalar , func_dict , force_active = False):

        """
        Helper function to apply a function with one scalar arg on target_kw.
        """
        data_type = target_kw.data_type
        if data_type in func_dict:
            func = func_dict[ data_type ]
            func( target_kw, scalar , force_active )
        else:
            raise Exception("scalar_apply_kw() only supported for INT/FLOAT/DOUBLE")

    def iadd_kw( self , target_kw , delta_kw , force_active = False):
        """
        The functions iadd_kw(), copy_kw(), set_kw(), scale_kw() and
        shift_kw() are not meant to be used as methods of the
        EclRegion class (altough that is of course perfectly OK) -
        rather a EclRegion instance is passed as an argument to an
        EclKW method, and then that method "flips things around" and
        calls one of these methods with the EclKW instance as
        argument. This applies to all the EclKW methods which take an
        optional "mask" argument.
        """
        if isinstance(delta_kw , EclKW):
            if target_kw.assert_binary( delta_kw ):
                self._iadd_kw( target_kw , delta_kw , force_active )
            else:
                raise TypeError("Type mismatch")
        else:
            self.shift_kw( target_kw , delta_kw , force_active = force_active)


    def shift_kw( self , ecl_kw , shift , force_active = False):
        """
        See usage documentation on iadd_kw().
        """
        self.scalar_apply_kw( ecl_kw , shift , {EclDataType.ECL_INT    : self._shift_kw_int,
                                                EclDataType.ECL_FLOAT  : self._shift_kw_float ,
                                                EclDataType.ECL_DOUBLE : self._shift_kw_double} , force_active)

    def isub_kw( self , target_kw , delta_kw , force_active = False):
        if isinstance(delta_kw , EclKW):
            if target_kw.assert_binary( delta_kw ):
                self._isub_kw( target_kw , delta_kw , force_active )
            else:
                raise TypeError("Type mismatch")
        else:
            self.shift_kw( target_kw , -delta_kw , force_active = force_active)


    def scale_kw( self , ecl_kw , scale , force_active = False):
        """
        See usage documentation on iadd_kw().
        """
        self.scalar_apply_kw( ecl_kw , scale , {EclDataType.ECL_INT    : self._scale_kw_int,
                                                EclDataType.ECL_FLOAT  : self._scale_kw_float ,
                                                EclDataType.ECL_DOUBLE : self._scale_kw_double} , force_active)

    def imul_kw(self, target_kw , other , force_active = False):
        if isinstance(other , EclKW):
            if target_kw.assert_binary( other):
                self._imul_kw( target_kw , other )
            else:
                raise TypeError("Type mismatch")
        else:
            self.scale_kw( target_kw , other , force_active )


    def idiv_kw( self , target_kw , other , force_active = False):
        if isinstance(other , EclKW):
            if target_kw.assert_binary( other):
                self._idiv_kw( target_kw , other )
            else:
                raise TypeError("Type mismatch")
        else:
            if target_kw.data_type.is_int():
                scale = 1 // other
            else:
                scale = 1.0 / other
            self.scale_kw( target_kw , scale , force_active )


    def copy_kw( self , target_kw , src_kw , force_active = False):
        """
        See usage documentation on iadd_kw().
        """
        if target_kw.assert_binary( src_kw ):
            self._copy_kw( target_kw , src_kw , force_active )
        else:
            raise TypeError("Type mismatch")

    def set_kw( self , ecl_kw , value , force_active = False):
        """
        See usage documentation on iadd_kw().
        """
        self.scalar_apply_kw( ecl_kw , value , {EclDataType.ECL_INT    : self._set_kw_int,
                                                EclDataType.ECL_FLOAT  : self._set_kw_float ,
                                                EclDataType.ECL_DOUBLE : self._set_kw_double} , force_active)


    def sum_kw(self, kw, force_active = False):
        data_type = kw.data_type
        if data_type == EclDataType.ECL_FLOAT:
            return self._sum_kw_float( kw, force_active )

        if data_type == EclDataType.ECL_INT:
            return self._sum_kw_int( kw, force_active )

        if data_type == EclDataType.ECL_DOUBLE:
            return self._sum_kw_double( kw, force_active )

        if data_type == EclDataType.ECL_BOOL:
            return self._sum_kw_bool( kw, force_active )

        raise ValueError("sum_kw only supported for; INT/FLOAT/DOUBLE/BOOL")


    #################################################################

    def ecl_region_instance(self):
        """
        Helper function (attribute) to support run-time typechecking.
        """
        return True


    def active_size(self):
        return len(self._get_active_list())


    def global_size(self):
        return len(self._get_global_list())


    def get_active_list(self):
        """
        IntVector instance with active indices in the region.
        """
        active_list = self._get_active_list()
        active_list.setParent(self)
        return active_list


    def get_global_list(self):
        """
        IntVector instance with global indices in the region.
        """
        global_list = self._get_global_list()
        global_list.setParent(self)
        return global_list


    def get_ijk_list(self):
        """
        WIll return a Python list of (ij,k) tuples for the region.
        """
        global_list = self.getGlobalList()
        ijk_list = []
        for g in global_list:
            ijk_list.append( self.grid.get_ijk( global_index = g ) )

        return ijk_list

    def contains_ijk( self , i,j,k):
        """
        Will check if the cell given by i,j,k is part of the region.
        """
        return self._contains_ijk( i , j , k )


    def contains_global( self , global_index):
        """
        Will check if the cell given by @global_index is part of the region.
        """
        return self._contains_global( global_index )


    def contains_active( self , active_index):
        """
        Will check if the cell given by @active_index is part of the region.
        """
        return self._contains_active( active_index )


    def kw_index_list(self , ecl_kw , force_active):
        c_ptr = self._get_kw_index_list( ecl_kw , force_active)
        index_list = IntVector.createCReference( c_ptr, self )
        return index_list

    @property
    def name(self):
        return self._get_name()

    def get_name(self):
        return self._get_name( )

    def set_name(self , name):
        self._set_name( name )

monkey_the_camel(EclRegion, 'selectTrue', EclRegion.select_true)
monkey_the_camel(EclRegion, 'selectFalse', EclRegion.select_false)
monkey_the_camel(EclRegion, 'selectFromLayer', EclRegion.select_from_layer)
monkey_the_camel(EclRegion, 'getActiveList', EclRegion.get_active_list)
monkey_the_camel(EclRegion, 'getGlobalList', EclRegion.get_global_list)
monkey_the_camel(EclRegion, 'getIJKList', EclRegion.get_ijk_list)
monkey_the_camel(EclRegion, 'getName', EclRegion.get_name)
monkey_the_camel(EclRegion, 'setName', EclRegion.set_name)



#  Copyright (C) 2013  Statoil ASA, Norway.
#
#  The file 'test_work_area.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
import os.path

from cwrap import BaseCClass
from ecl import EclPrototype


class TestArea(BaseCClass):
    _test_area_alloc           = EclPrototype("void* test_work_area_alloc( char* )" , bind = False)
    _test_area_alloc_relative  = EclPrototype("void* test_work_area_alloc_relative( char* , char* )" , bind = False)
    _free                      = EclPrototype("void test_work_area_free( test_area )")
    _install_file              = EclPrototype("void test_work_area_install_file( test_area , char* )")
    _copy_directory            = EclPrototype("void test_work_area_copy_directory( test_area , char* )")
    _copy_file                 = EclPrototype("void test_work_area_copy_file( test_area , char* )")
    _copy_directory_content    = EclPrototype("void test_work_area_copy_directory_content( test_area , char* )")
    _copy_parent_directory     = EclPrototype("void test_work_area_copy_parent_directory( test_area , char* )")
    _copy_parent_content       = EclPrototype("void test_work_area_copy_parent_content( test_area , char* )")
    _get_cwd                   = EclPrototype("char* test_work_area_get_cwd( test_area )")
    _get_original_cwd          = EclPrototype("char* test_work_area_get_original_cwd( test_area )")
    _set_store                 = EclPrototype("void test_work_area_set_store( test_area , bool)")
    _sync                      = EclPrototype("void test_work_area_sync( test_area )")

    def __init__(self, test_name, prefix = None , store_area=False , c_ptr = None):

        if c_ptr is None:
            if prefix:
                if os.path.exists( prefix ):
                    c_ptr = self._test_area_alloc_relative(prefix , test_name)
                else:
                    raise IOError("The prefix path:%s must exist" % prefix)
            else:
                c_ptr = self._test_area_alloc(test_name)

        super(TestArea, self).__init__(c_ptr)
        self.set_store( store_area )


    def get_original_cwd(self):
        return self._get_original_cwd()

    def get_cwd(self):
        return self._get_cwd()

    def orgPath(self , path):
        if os.path.isabs( path ):
            return path
        else:
            return os.path.abspath( os.path.join( self.get_original_cwd( ) , path ) )


    # All the methods install_file() , copy_directory(),
    # copy_parent_directory(), copy_parent_content(),
    # copy_directory_content() and copy_file() expect an input
    # argument which is relative to the original CWD - or absolute.

    def install_file( self, filename):
        if os.path.isfile(self.orgPath(filename)):
            self._install_file(filename)
        else:
            raise IOError("No such file:%s" % filename)


    def copy_directory( self, directory):
        if os.path.isdir( self.orgPath(directory) ):
            self._copy_directory(directory)
        else:
            raise IOError("No such directory: %s" % directory)

    def copy_parent_directory( self , path):
        if os.path.exists( self.orgPath(path) ):
            self._copy_parent_directory( path)
        else:
            raise IOError("No such file or directory: %s" % path)


    def copy_parent_content( self , path):
        if os.path.exists( self.orgPath(path) ):
            self._copy_parent_content(path)
        else:
            raise IOError("No such file or directory: %s" % path)

    def copy_directory_content( self, directory):
        if os.path.isdir( self.orgPath(directory) ):
            self._copy_directory_content(directory)
        else:
            raise IOError("No such directory: %s" % directory )


    def copy_file( self, filename):
        if os.path.isfile( self.orgPath(filename) ):
            self._copy_file(filename)
        else:
            raise IOError("No such file:%s" % filename)


    def free(self):
        self._free()


    def set_store(self, store):
        self._set_store(store)


    def getFullPath(self , path):
        if not os.path.exists( path ):
            raise IOError("Path not found:%s" % path)

        if os.path.isabs( path ):
            raise IOError("Path:%s is already absolute" % path)

        return os.path.join( self.get_cwd() , path )


    def sync(self):
        return self._sync( )



class TestAreaContext(object):
    def __init__(self, test_name, prefix = None , store_area=False):
        self.test_name = test_name
        self.store_area = store_area
        self.prefix = prefix

    def __enter__(self):
        """
         @rtype: TestArea
        """
        self.test_area = TestArea(self.test_name, prefix = self.prefix , store_area = self.store_area )
        return self.test_area


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.test_area.free()      # free the TestData object (and cd back to the original dir)
        self.test_area.free = None # avoid double free
        del self.test_area
        return False



#!/usr/bin/env python
#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

class Cell(object):

    def __init__(self, grid, global_index):
        self._grid = grid
        self._idx = global_index
        self._ijk = grid.get_ijk(global_index=self._idx)
        self._aidx = grid.get_active_index(global_index=self._idx)

    @property
    def volume(self):
        return self._grid.cell_volume(global_index=self._idx)

    @property
    def global_index(self):
        return self._idx

    @property
    def active_index(self):
        return self._aidx

    @property
    def ijk(self):
        return self._ijk
    @property
    def i(self):
        return self._ijk[0]
    @property
    def j(self):
        return self._ijk[1]
    @property
    def k(self):
        return self._ijk[2]

    @property
    def active(self):
        return self._aidx >= 0

    def eval(self, kw):
        return self._grid.grid_value(kw, self.i, self.j, self.k)

    @property
    def fracture(self):
        return self._grid.active_fracture_index(global_index=self._idx)

    @property
    def dz(self):
        return self._grid.cell_dz(global_index=self._idx)

    @property
    def dimension(self):
        return self._grid.get_cell_dims(global_index=self._idx)

    @property
    def valid_geometry(self):
        return self._grid.valid_cell_geometry(global_index=self._idx)

    @property
    def valid(self):
        return not self._grid.cell_invalid(global_index=self._idx)

    def __contains__(self, coord):
        """
        Will check if this cell contains point given by world
        coordinates (x,y,z)=coord.
        """
        if len(coord) != 3:
            raise ValueError('Cell contains takes a triple (x,y,z), was given %s.' % coord)
        x,y,z = coord
        return self._grid._cell_contains(self._idx, x,y,z)

    def __eq__(self, other):
        if isinstance(other, Cell):
            idx_eq = self.global_index == other.global_index
            return idx_eq and self._grid == other._grid
        return NotImplemented

    def __neq__(self, other):
        if isinstance(other, Cell):
            return not self == other
        return NotImplemented

    def hash(self):
        return hash((self._idx, self._aidx, self.ijk))

    @property
    def coordinate(self):
        return self._grid.get_xyz(global_index=self._idx)

    @property
    def corners(self):
        """
        Return xyz for each of the eight vertices, indexed by:

        lower layer:   upper layer

         2---3           6---7
         |   |           |   |
         0---1           4---5
        """
        cs = lambda c : self._grid.get_cell_corner(c, global_index=self._idx)
        return [cs(i) for i in range(8)]

    def __repr__(self):
        act = 'active' if self.active else 'inactive'
        pos = '(%.3f, %.3f, %.3f)' % self.coordinate
        cnt = '%d, %d, %d, %s, %s, grid=%s' % (self.i, self.j, self.k,
                                               act, pos,
                                               self._grid.get_name())

        return 'Cell(%s)' % cnt



#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from os.path import isfile
from cwrap import BaseCClass
from ecl.grid import EclGrid
from ecl.eclfile.ecl_file import EclFile
from ecl.well import WellTimeLine
from ecl import EclPrototype


class WellInfo(BaseCClass):
    TYPE_NAME = "well_info"

    _alloc            = EclPrototype("void* well_info_alloc(ecl_grid)", bind = False)
    _free             = EclPrototype("void  well_info_free(well_info)")
    _load_rstfile     = EclPrototype("void  well_info_load_rstfile(well_info, char*, bool)")
    _load_rst_eclfile = EclPrototype("void  well_info_load_rst_eclfile(well_info, ecl_file, bool)")
    _get_well_count   = EclPrototype("int   well_info_get_num_wells(well_info)")
    _iget_well_name   = EclPrototype("char* well_info_iget_well_name(well_info, int)")
    _has_well         = EclPrototype("bool  well_info_has_well(well_info, char*)")
    _get_ts           = EclPrototype("well_time_line_ref well_info_get_ts(well_info, char*)")


    def __init__(self, grid, rst_file=None, load_segment_information=True):
        """
        @type grid: EclGrid
        @type rst_file: str or EclFile or list of str or list of EclFile
        """
        c_ptr = self._alloc(grid)
        super(WellInfo, self).__init__(c_ptr)
        if not c_ptr:
            raise ValueError('Unable to construct WellInfo from grid %s.' % str(grid))

        if rst_file is not None:
            if isinstance(rst_file, list):
                for item in rst_file:
                    self.addWellFile(item, load_segment_information)
            else:
                self.addWellFile(rst_file, load_segment_information)


    def __repr__(self):
        return 'WellInfo(well_count = %d) at 0x%x' % (len(self), self._address())

    def __len__(self):
        """ @rtype: int """
        return self._get_well_count( )


    def __getitem__(self, item):
        """
         @type item: int or str
         @rtype: WellTimeLine
        """

        if isinstance(item, str):
            if not item in self:
                raise KeyError("The well '%s' is not in this set." % item)
            well_name = item

        elif isinstance(item, int):
            if not 0 <= item < len(self):
                raise IndexError("Index must be in range 0 <= %d < %d" % (item, len(self)))
            well_name = self._iget_well_name( item )

        return self._get_ts(well_name).setParent(self)

    def __iter__(self):
        """ @rtype: iterator of WellTimeLine """
        index = 0

        while index < len(self):
            yield self[index]
            index += 1


    def allWellNames(self):
        """ @rtype: list of str """
        return [self._iget_well_name(index) for index in range(0, len(self))]


    def __contains__(self, item):
        """
         @type item: str
         @rtype: bool
        """
        return self._has_well( item )

    def _assert_file_exists(self, rst_file):
        if not isfile(rst_file):
            raise IOError('No such file %s' % rst_file)

    def addWellFile(self, rst_file, load_segment_information):
        """ @type rstfile: str or EclFile """
        if isinstance(rst_file, str):
            self._assert_file_exists(rst_file)
            self._load_rstfile(rst_file, load_segment_information)
        elif isinstance(rst_file, EclFile):
            self._load_rst_eclfile(rst_file, load_segment_information)
        else:
            raise TypeError("Expected the RST file to be a filename or an EclFile instance.")


    def hasWell(self , well_name):
        return well_name in self


    def free(self):
        self._free( )



#  Copyright (C) 2011 Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Create a polygon
"""
import ctypes
import os.path

from cwrap import BaseCClass
from ecl import EclPrototype
from .geometry_tools import GeometryTools


class CPolyline(BaseCClass):
    TYPE_NAME = "geo_polygon"

    _alloc_new          = EclPrototype("void*           geo_polygon_alloc( char* )" , bind = False)
    _fread_alloc_irap   = EclPrototype("geo_polygon_obj geo_polygon_fload_alloc_irap( char* )" , bind = False)
    _add_point          = EclPrototype("void     geo_polygon_add_point( geo_polygon , double , double )")
    _add_point_front    = EclPrototype("void     geo_polygon_add_point_front( geo_polygon , double , double )")
    _free               = EclPrototype("void     geo_polygon_free( geo_polygon )")
    _size               = EclPrototype("int      geo_polygon_get_size( geo_polygon )")
    _iget_xy            = EclPrototype("void     geo_polygon_iget_xy( geo_polygon , int , double* , double* )")
    _segment_intersects = EclPrototype("bool     geo_polygon_segment_intersects( geo_polygon , double , double, double , double)")
    _get_name           = EclPrototype("char*    geo_polygon_get_name( geo_polygon  )")
    _set_name           = EclPrototype("void     geo_polygon_set_name( geo_polygon , char*  )")
    _segment_length     = EclPrototype("double   geo_polygon_get_length( geo_polygon)")
    _equal              = EclPrototype("bool     geo_polygon_equal( geo_polygon , geo_polygon )")


    def __init__(self, name = None , init_points = ()):
        c_ptr = self._alloc_new( name )
        super(CPolyline , self).__init__( c_ptr )
        for (xc, yc) in init_points:
            self.addPoint(xc, yc)


    @classmethod
    def createFromXYZFile(cls , filename , name = None):
        if not os.path.isfile(filename):
            raise IOError("No such file:%s" % filename)

        polyline = cls._fread_alloc_irap( filename )
        if not name is None:
            polyline._set_name( name )
        return polyline

    def __str__(self):
        name = self.getName()
        if name:
            str = "%s [" % name
        else:
            str = "["

        for index,p in enumerate(self):
            str += "(%g,%g)" % p
            if index < len(self) - 1:
                str += ","
        str += "]"
        return str

    def __repr__(self):
        return str(self)


    def __len__(self):
        return self._size()


    def __getitem__(self , index):
        if not isinstance(index,int):
            raise TypeError("Index argument must be integer. Index:%s invalid" % index)

        if index < 0:
            index += len(self)

        if 0 <= index < len(self):
            x = ctypes.c_double()
            y = ctypes.c_double()
            self._iget_xy( index , ctypes.byref(x) , ctypes.byref(y) )

            return (x.value , y.value)
        else:
            raise IndexError("Invalid index:%d valid range: [0,%d)" % (index , len(self)))


    def segmentIntersects(self, p1 , p2):
        return self._segment_intersects(p1[0] , p1[1] , p2[0] , p2[1])


    def intersects(self , polyline):
        if len(self) > 1:
            for index,p2 in enumerate(polyline):
                if index == 0:
                    continue

                p1 = polyline[index - 1]
                if self.segmentIntersects(p1 , p2):
                    return True
        return False


    def __iadd__(self , other ):
        for p in other:
            self.addPoint( p[0] , p[1] )
        return self


    def __add__(self , other ):
        copy = CPolyline( init_points = self)
        copy.__iadd__(other)
        return copy


    def __radd__(self , other ):
        copy = CPolyline( init_points = other )
        copy.__iadd__(self)
        return copy

    def __eq__(self , other):
        if super(CPolyline , self).__eq__( other ):
            return True
        else:
            return self._equal( other )


    def segmentLength(self):
        if len(self) == 0:
            raise ValueError("Can not measure length of zero point polyline")

        return self._segment_length( )

    def extendToBBox(self , bbox , start = True):
        if start:
            p0 = self[1]
            p1 = self[0]
        else:
            p0 = self[-2]
            p1 = self[-1]

        ray_dir = GeometryTools.lineToRay(p0,p1)
        intersections = GeometryTools.rayPolygonIntersections( p1 , ray_dir , bbox)
        if intersections:
            p2 = intersections[0][1]
            if self.getName():
                name = "Extend:%s" % self.getName()
            else:
                name = None

            return CPolyline( name = name , init_points = [(p1[0] , p1[1]) , p2])
        else:
            raise ValueError("Logical error - must intersect with bounding box")


    def addPoint( self, xc, yc , front = False):
        if front:
            self._add_point_front(xc, yc)
        else:
            self._add_point(xc, yc)


    def getName(self):
        return self._get_name( )


    def free(self):
        self._free( )


    def unzip(self):
        x_list = [ ]
        y_list = [ ]
        for x,y in self:
            x_list.append(x)
            y_list.append(y)

        return (x_list , y_list)


    def unzip2(self):
        return self.unzip()


    def connect(self , target):
        end1 = self[0]
        end2 = self[-1]

        p1 = GeometryTools.nearestPointOnPolyline( end1 , target )
        p2 = GeometryTools.nearestPointOnPolyline( end2 , target )

        d1 = GeometryTools.distance( p1 , end1 )
        d2 = GeometryTools.distance( p2 , end2 )

        if d1 < d2:
            return [end1 , p1]
        else:
            return [end2 , p2]



#  Copyright (C) 2014.  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from __future__ import print_function

from ecl.util.util import monkey_the_camel


class FaultSegment(object):

    def __init__(self, C1, C2):
        self.__C1 = C1
        self.__C2 = C2
        self.__next_segment = None


    def __eq__(self, other):
        s = self.c1, self.c2
        o = other.c1, other.c2
        o_flipped = other.c2, other.c1
        return s == o or s == o_flipped

    def __hash__(self):
        return hash(hash(self.__C1) + hash(self.__C2) + hash(self.__next_segment))

    def get_corners(self):
        return (self.__C1, self.__C2)

    def joins(self, other):
        if self.__C1 == other.__C1:
            return True
        if self.__C1 == other.__C2:
            return True
        if self.__C2 == other.__C1:
            return True
        if self.__C2 == other.__C2:
            return True

        return False

    def get_c1(self):
        return self.__C1

    @property
    def c1(self):
        return self.__C1

    def get_c2(self):
        return self.__C2

    @property
    def c2(self):
        return self.__C2

    def swap(self):
        C1 = self.__C1
        self.__C1 = self.__C2
        self.__C2 = C1


    def __repr__(self):
        return "%d -> %d" % (self.__C1, self.__C2)



class SegmentMap(object):

    def __init__(self):
        self.__segment_map = {}
        self.__count_map = {}

    def __len__(self):
        return len(self.__segment_map)

    def __str__(self):
        return self.__segment_map.__str__()

    def verify(self):
        for (C, count) in self.__count_map.iteritems():
            if count > 0:
                d = self.__segment_map[C]
                if len(d) != count:
                    print('CornerPoint:%d  count:%d  len(d):%d map:%s' % (C, count, len(d), d))
                assert len(d) == count
            else:
                assert self.__segment_map.get(C) is None


    def add_segment(self, segment):
        (C1,C2) = segment.getCorners()
        if C1 not in self.__segment_map:
            self.__segment_map[C1] = {}
            self.__count_map[C1] = 0
        if C2 not in self.__segment_map:
            self.__segment_map[C2] = {}
            self.__count_map[C2] = 0

        if C2 not in self.__segment_map[C1]:
            self.__segment_map[C1][C2] = segment
            self.__count_map[C1] += 1

        if C1 not in self.__segment_map[C2]:
            self.__segment_map[C2][C1] = segment
            self.__count_map[C2] += 1



    def del_segment(self, segment):
        (C1,C2) = segment.getCorners()
        self.__count_map[C1] -= 1
        self.__count_map[C2] -= 1
        del self.__segment_map[C1][C2]
        del self.__segment_map[C2][C1]

        if len(self.__segment_map[C1]) == 0:
            del self.__segment_map[C1]

        if len(self.__segment_map[C2]) == 0:
            del self.__segment_map[C2]


    def pop_start(self):
        end_segments = []
        for (C, count) in self.__count_map.items():
            if count == 1:
                end_segments.append(list(self.__segment_map[C].values())[0])

        start_segment = end_segments[0]
        self.delSegment(start_segment)
        return start_segment

    def pop_next(self, segment):
        (C1,C2) = segment.getCorners()
        if self.__count_map[C1] >= 1:
            next_segment = list(self.__segment_map[C1].values())[0]
        elif self.__count_map[C2] >= 1:
            next_segment = list(self.__segment_map[C2].values())[0]
        else:
            next_segment = None

        if next_segment:
            self.delSegment(next_segment)
        return next_segment


    def print_content(self):
        for d in self.__segment_map.values():
            for (C,S) in d.iteritems():
                print(S)



monkey_the_camel(FaultSegment, 'getCorners', FaultSegment.get_corners)
monkey_the_camel(FaultSegment, 'getC1', FaultSegment.get_c1)
monkey_the_camel(FaultSegment, 'getC2', FaultSegment.get_c2)

monkey_the_camel(SegmentMap, 'addSegment', SegmentMap.add_segment)
monkey_the_camel(SegmentMap, 'delSegment', SegmentMap.del_segment)
monkey_the_camel(SegmentMap, 'popStart', SegmentMap.pop_start)
monkey_the_camel(SegmentMap, 'popNext', SegmentMap.pop_next)
monkey_the_camel(SegmentMap, 'printContent', SegmentMap.print_content)



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'rng.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
import os.path

from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.util.enums import RngInitModeEnum, RngAlgTypeEnum


class RandomNumberGenerator(BaseCClass):
    TYPE_NAME = "rng"

    _rng_alloc = EclPrototype("void* rng_alloc(rng_alg_type_enum, rng_init_mode_enum)" , bind = False)
    _free = EclPrototype("void rng_free(rng)")
    _get_double = EclPrototype("double rng_get_double(rng)")
    _get_int = EclPrototype("int rng_get_int(rng, int)")
    _get_max_int = EclPrototype("uint rng_get_max_int(rng)")
    _state_size = EclPrototype("int rng_state_size(rng)")
    _set_state = EclPrototype("void rng_set_state(rng , char*)")
    _load_state = EclPrototype("void rng_load_state(rng , char*)")
    _save_state = EclPrototype("void rng_save_state(rng , char*)")

    def __init__(self, alg_type=RngAlgTypeEnum.MZRAN, init_mode=RngInitModeEnum.INIT_CLOCK):
        assert isinstance(alg_type, RngAlgTypeEnum)
        assert isinstance(init_mode, RngInitModeEnum)

        c_ptr = self._rng_alloc(alg_type, init_mode)
        super(RandomNumberGenerator, self).__init__(c_ptr)

    def stateSize(self):
        return self._state_size()

    def setState(self, seed_string):
        state_size = self.stateSize()
        if len(seed_string) < state_size:
            raise ValueError("The seed string must be at least %d characters long" % self.stateSize())
        self._set_state( seed_string)

    def getDouble(self):
        """ @rtype: float """
        return self._get_double()

    def getInt(self, max=None):
        """ @rtype: float """
        if max is None:
            max = self._get_max_int()

        return self._get_int(max)

    def free(self):
        self._free()

    def loadState(self , seed_file):
        """
        Will seed the RNG from the file @seed_file.
        """
        if os.path.isfile( seed_file ):
            self._load_state( seed_file )
        else:
            raise IOError("No such file: %s" % seed_file)


    def saveState(self , seed_file):
        """
        Will save the state of the rng to @seed_file
        """
        self._save_state( seed_file )
        



class EclSumNode(object):

    def __init__(self, report_step, days, date, mpl_date, value):
        """
        EclSumNode is a 'struct' with a summary value and time.

        EclSumNode - a small 'struct' with a summary value and time in
        several formats. When iterating over a EclSumVector instance
        you will get EclSumNode instances. The content of the
        EclSumNode type is stored as plain attributes:

            value       : The actual value
            report_step : The report step
            days        : Days since simulation start
            date        : The simulation date
            mpl_date    : A date format suitable for matplotlib

        """
        self.value = value
        self.report_step = report_step
        self.days = days
        self.date = date
        self.mpl_date = mpl_date

    def __repr__(self):
        return "EclSumNode(days=%d, value=%g)" % (self.days, self.value)



#  Copyright (C) 2011  Statoil ASA, Norway. 
#   
#  The file 'lookup_table.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details. 


from cwrap import BaseCClass
from ecl import EclPrototype


class LookupTable(BaseCClass):
    _alloc =          EclPrototype("void* lookup_table_alloc_empty()" , bind = False)
    _max =            EclPrototype("double lookup_table_get_max_value( lookup_table )")
    _min =            EclPrototype("double lookup_table_get_min_value( lookup_table )")
    _arg_max =        EclPrototype("double lookup_table_get_max_arg( lookup_table )")
    _arg_min =        EclPrototype("double lookup_table_get_min_arg( lookup_table )")
    _append =         EclPrototype("void lookup_table_append( lookup_table , double , double )")
    _size =           EclPrototype("int lookup_table_get_size( lookup_table )")
    _interp =         EclPrototype("double lookup_table_interp( lookup_table , double)")
    _free =           EclPrototype("void lookup_table_free( lookup_table )")
    _set_low_limit =  EclPrototype("void lookup_table_set_low_limit( lookup_table , double)")
    _set_high_limit = EclPrototype("void lookup_table_set_high_limit( lookup_table , double)")
    _has_low_limit =  EclPrototype("bool lookup_table_has_low_limit( lookup_table)")
    _has_high_limit = EclPrototype("bool lookup_table_has_high_limit( lookup_table)")

    def __init__(self, lower_limit=None, upper_limit=None):
        super(LookupTable, self).__init__(self._alloc())

        if not lower_limit is None:
            self.setLowerLimit(lower_limit)

        if not upper_limit is None:
            self.setUpperLimit(upper_limit)

    def getMaxValue(self):
        self.assertSize(1)
        return self._max()

    def getMinValue(self):
        self.assertSize(1)
        return self._min()

    def getMinArg(self):
        self.assertSize(1)
        return self._arg_min()

    def getMaxArg(self):
        self.assertSize(1)
        return self._arg_max()

    def assertSize(self, N):
        if len(self) < N:
            raise ValueError("Lookup table is too small")

    def __len__(self):
        return self._size()

    @property
    def size(self):
        return len(self)

    # Deprecated properties
    @property
    def max(self):
        return self.getMaxValue()

    @property
    def min(self):
        return self.getMinValue()

    @property
    def arg_max(self):
        return self.getMaxArg()

    @property
    def arg_min(self):
        return self.getMinArg()

    def setLowerLimit(self, value):
        self._set_low_limit(value)

    def hasLowerLimit(self):
        return self._has_low_limit()

    def setUpperLimit(self, value):
        self._set_high_limit(value)

    def hasUpperLimit(self):
        return self._has_high_limit()

    def interp(self, x):
        self.assertSize(2)
        if x < self.getMinArg():
            if not self.hasLowerLimit():
                raise ValueError("Interpolate argument:%g is outside valid interval: [%g,%g]" % (x, self.getMinArg(), self.getMaxArg()))
        elif x > self.getMaxArg():
            if not self.hasUpperLimit():
                raise ValueError("Interpolate argument:%g is outside valid interval: [%g,%g]" % (x, self.getMinArg(), self.getMaxArg()))

        return self._interp(x)

    def append(self, x, y):
        self._append( x, y)

    #todo: necessary???
    def __del__(self):
        self._free()

    def free(self):
        self._free( )




#  Copyright (C) 2018  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

"""
  ecl_rft/[EclRFTFile , EclRFT , EclRFTCell]: Loads an ECLIPSE RFT/PLT
     file, and can afterwords be used to support various queries.
"""

from .well_trajectory import WellTrajectory
from .ecl_rft_cell import EclPLTCell, EclRFTCell
from .ecl_rft import EclRFT, EclRFTFile




#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  The file 'ecl_grid_generator.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

import itertools, numpy
from math import sqrt

from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
from ecl.util.util import IntVector
from ecl import EclDataType
from ecl.eclfile import EclKW
from ecl.grid import EclGrid

def flatten(l):
    return [elem for sublist in l for elem in sublist]

def divide(l, size):
    return [l[i:i+size:] for i in range(0, len(l), size)]

def duplicate_inner(l):
    return [elem for elem in l for i in range(2)][1:-1:]

def construct_floatKW(name, values):
    kw = EclKW(name, len(values), EclDataType.ECL_FLOAT)
    for i, value in enumerate(values):
        kw[i] = value
    return kw

def pre_mapaxes_translation(translation, mapaxes):
    if mapaxes is None:
        return translation

    x, y, z = translation

    unit_y = numpy.array((mapaxes[0]-mapaxes[2], mapaxes[1]-mapaxes[3]));
    unit_y /= sqrt(numpy.sum(unit_y*unit_y))

    unit_x = numpy.array((mapaxes[4]-mapaxes[2], mapaxes[5]-mapaxes[3]));
    unit_x /= sqrt(numpy.sum(unit_x*unit_x))

    det = 1.0 / (unit_x[0]*unit_y[1] - unit_x[1] * unit_y[0]);

    return (
                ( x*unit_y[1] - y*unit_y[0]) * det,
                (-x*unit_x[1] + y*unit_x[0]) * det,
                z
           )

class EclGridGenerator:

    _alloc_rectangular = EclPrototype("ecl_grid_obj ecl_grid_alloc_rectangular(int, int, int, double, double, double, int*)", bind=False)

    @classmethod
    def create_rectangular(cls, dims, dV, actnum=None):
        """
        Will create a new rectangular grid. @dims = (nx,ny,nz)  @dVg = (dx,dy,dz)

        With the default value @actnum == None all cells will be active,
        """
        if actnum is None:
            ecl_grid = cls._alloc_rectangular(
                                     dims[0], dims[1], dims[2],
                                     dV[0], dV[1], dV[2],
                                     None
                                     )
        else:
            if not isinstance(actnum, IntVector):
                tmp = IntVector(initial_size=len(actnum))
                for (index, value) in enumerate(actnum):
                    tmp[index] = value
                actnum = tmp

            if not len(actnum) == dims[0]*dims[1]*dims[2]:
                raise ValueError(
                        "ACTNUM size mismatch: len(ACTNUM):%d  Expected:%d"
                        % (len(actnum), dims[0]*dims[1]*dims[2])
                        )

            ecl_grid = cls._alloc_rectangular(
                                 dims[0], dims[1], dims[2],
                                 dV[0], dV[1], dV[2],
                                 actnum.getDataPtr()
                                 )

        # If we have not succeeded in creatin the grid we *assume* the
        # error is due to a failed malloc.
        if ecl_grid is None:
            raise MemoryError("Failed to allocated regualar grid")

        return ecl_grid

    @classmethod
    def create_single_cell_grid(cls, corners):
        """
        Provided with the corners of the grid in a similar manner as the eight
        corners are output for a single cell, this method will create a grid
        consisting of a single cell with the specified corners as its corners.
        """

        zcorn = [corners[i][2] for i in range(8)]

        coord = [(corners[i], corners[i+4]) for i in range(4)]
        coord = flatten(flatten(coord))

        def construct_floatKW(name, values):
            kw = EclKW(name, len(values), EclDataType.ECL_FLOAT)
            for i in range(len(values)):
                kw[i] = values[i]
            return kw

        grid = EclGrid.create(
                (1, 1, 1),
                construct_floatKW("ZCORN", zcorn),
                construct_floatKW("COORD", coord),
                None
                )

        if not corners == [grid.getCellCorner(i, 0) for i in range(8)]:
            raise AssertionError("Failed to generate single cell grid. " +
                    "Did not end up the expected corners.")

        return grid

    @classmethod
    def create_zcorn(cls, dims, dV, offset=1, escape_origo_shift=(1,1,0),
            irregular_offset=False, irregular=False, concave=False,
            faults=False):

        cls.__assert_zcorn_parameters(dims, dV, offset, escape_origo_shift,
                    irregular_offset, irregular, concave, faults)

        nx, ny, nz = dims
        dx, dy, dz = dV

        # Compute zcorn
        z = escape_origo_shift[2]
        zcorn = [z]*(4*nx*ny)
        for k in range(nz-1):
            z = z+dz
            local_offset = offset + (dz/2. if irregular_offset and k%2 == 0 else 0)

            layer = []
            for i in range(ny+1):
                shift = ((i if concave else 0) + (k/2 if irregular else 0)) % 2
                path = [z if i%2 == shift else z+local_offset for i in range(nx+1)]
                layer.append(duplicate_inner(path))

            zcorn = zcorn + (2*flatten(duplicate_inner(layer)))

        z = z+dz
        zcorn = zcorn + ([z]*(4*nx*ny))

        if faults:
            # Ensure that drop does not align with grid structure
            drop = (offset+dz)/2. if abs(offset-dz/2.) > 0.2 else offset + 0.4
            zcorn = cls.__create_faults(nx, ny, nz, zcorn, drop)


        if z != escape_origo_shift[2] + nz*dz:
            raise ValueError("%f != %f" % (z, escape_origo_shift[2] + nz*dz))

        cls.assert_zcorn(nx, ny, nz, zcorn)
        return construct_floatKW("ZCORN", zcorn)

    @classmethod
    def create_coord(cls, dims, dV, escape_origo_shift=(1,1,0),
            scale=1, translation=(0,0,0), rotate=False, misalign=False):

        nx, ny, nz = dims
        dx, dy, dz = dV

        # Compute coord
        z = escape_origo_shift[2] + nz*dz
        coord = []
        for j, i in itertools.product(range(ny+1), range(nx+1)):
            x, y = i*dx+escape_origo_shift[0], j*dy+escape_origo_shift[1]
            coord = coord + [x, y, escape_origo_shift[2], x, y, z]

        # Apply transformations
        lower_center = (
                nx*dx/2. + escape_origo_shift[0],
                ny*dy/2. + escape_origo_shift[1]
                )

        if misalign:
            coord = cls.__misalign_coord(coord, dims, dV)

        coord = cls.__scale_coord(coord, scale, lower_center)

        if rotate:
            coord = cls.__rotate_coord(coord, lower_center)

        coord = cls.__translate_lower_coord(coord, translation)

        cls.assert_coord(nx, ny, nz, coord)
        return construct_floatKW("COORD", coord)

    @classmethod
    def __assert_zcorn_parameters(cls, dims, dV, offset, escape_origo_shift,
            irregular_offset, irregular, concave, faults):

        nx, ny, nz = dims
        dx, dy, dz = dV

        # Validate arguments
        if min(dims + dV) <= 0:
            raise ValueError("Expected positive grid and cell dimentions")

        if offset < 0:
            raise ValueError("Expected non-negative offset")

        if irregular and offset + (dz/2. if irregular_offset else 0) > dz:
            raise AssertionError("Arguments can result in self-" +
                    "intersecting cells. Increase dz, deactivate eiter " +
                    "irregular or irregular_offset, or decrease offset to avoid " +
                    "any problems")

    @classmethod
    def create_grid(cls, dims, dV, offset=1,
            escape_origo_shift=(1,1,0),
            irregular_offset=False, irregular=False, concave=False,
            faults=False, scale=1, translation=(0,0,0), rotate=False,
            misalign=False):
        """
        Will create a new grid where each cell is a parallelogram (skewed by z-value).
        The number of cells are given by @dims = (nx, ny, nz) and the dimention
        of each cell by @dV = (dx, dy, dz).

        All cells are guaranteed to not be self-intersecting. Hence, no twisted
        cells and somewhat meaningfull cells.

        @offset gives how much the layers should fluctuate or "wave" as you
        move along the X-axis.

        @irregular_offset decides whether the offset should be constant or
        increase by dz/2 every now and then.

        @irregular if true some of the layers will be inclining and others
        declining at the start.

        @concave decides whether the cells are to be convex or not. In
        particular, if set to False, all cells of the grid will be concave.

        @escape_origo_shift is used to prevent any cell of having corners in (0,0,z)
        as there is a heuristic in ecl_grid.c that marks such cells as tainted.

        @faults decides if there are to be faults in the grid.

        @scale A positive number that scales the "lower" endpoint of all
        coord's. In particular, @scale != 1 creates trapeziod cells in both the XZ
        and YZ-plane.

        @translation the lower part of the grid is translated ("slided") by the specified
        additive factor.

        @rotate the lower part of the grid is rotated 90 degrees around its
        center.

        @misalign will toggle COORD's slightly in various directions to break
        alignment

        Note that cells in the lowermost layer can have multiple corners
        at the same point.

        For testing it should give good coverage of the various scenarios this
        method can produce, by leting @dims be (10,10,10), @dV=(2,2,2), @offset=1,
        and try all 4 different configurations of @concave and
        @irregular_offset.
        """

        zcorn = cls.create_zcorn(dims, dV, offset, escape_origo_shift,
                                irregular_offset, irregular, concave, faults)

        coord = cls.create_coord(dims, dV, escape_origo_shift, scale,
                                translation, rotate, misalign)

        return EclGrid.create(dims, zcorn, coord, None)

    @classmethod
    def __create_faults(cls, nx, ny, nz, zcorn, drop):
        """
        Will create several faults consisting of all cells such that either its
        i or j index is 1 modulo 3.
        """

        plane_size = 4*nx*ny
        for x, y, z in itertools.product(range(nx), range(ny), range(nz)):
            if x%3 != 1 and y%3 != 1:
                continue

            corner = [0]*8
            corner[0] = 2*z*plane_size + 4*y*nx + 2*x
            corner[1] = corner[0] + 1
            corner[2] = corner[0] + 2*nx
            corner[3] = corner[2] + 1

            for i in range(4, 8):
                corner[i] = corner[i-4] + plane_size

            for c in corner:
                zcorn[c] = zcorn[c] + drop

        return zcorn

    @classmethod
    def assert_zcorn(cls, nx, ny, nz, zcorn, twisted_check=True):
        """

        Raises an AssertionError if the zcorn is not as expected. In
        patricular, it is verified that:

            - zcorn has the approperiate length (8*nx*ny*nz) and
            - that no cell is twisted.

        """

        if len(zcorn) != 8*nx*ny*nz:
            raise AssertionError(
                    "Expected len(zcorn) to be %d, was %d" %
                    (8*nx*ny*nz, len(zcorn))
                    )

        plane_size = 4*nx*ny
        for p in range(8*nx*ny*nz - plane_size):
            if zcorn[p] > zcorn[p + plane_size] and twisted_check:
                raise AssertionError(
                    "Twisted cell was created. " +
                    "Decrease offset or increase dz to avoid this!"
                    )

    @classmethod
    def __scale_coord(cls, coord, scale, lower_center):
        coord = numpy.array([
            list(map(float, coord[i:i+6:]))
            for i in range(0, len(coord), 6)
            ])
        origo = numpy.array(3*[0.] + list(lower_center) + [0])
        scale = numpy.array(3*[1.] + 2*[scale] + [1])
        coord = scale * (coord-origo) + origo
        return coord.flatten().tolist()

    @classmethod
    def __misalign_coord(cls, coord, dims, dV):
        nx, ny, nz = dims

        coord = numpy.array([
            list(map(float, coord[i:i+6:]))
            for i in range(0, len(coord), 6)
            ])

        tf = lambda i, j: 1./2 if abs(i)+abs(j) <= 1 else 0.25
        adjustment = numpy.array([
            (0, 0, 0, i*tf(i,j)*dV[0], j*tf(i,j)*dV[1], 0) for i, j in itertools.product([-1, 0, 1], repeat=2)
            ])

        for i, c in enumerate(coord):
            # Leave the outermost coords alone
            if i < nx+1 or i >= len(coord)-(nx+1):
                continue
            if i%(nx+1) in [0, nx]:
                continue

            c += adjustment[i%len(adjustment)]

        return coord.flatten().tolist()

    @classmethod
    def __rotate_coord(cls, coord, lower_center):
        coord = numpy.array([
            list(map(float, coord[i:i+6:]))
            for i in range(0, len(coord), 6)
            ])

        origo = numpy.array(3*[0.] + list(lower_center) + [0])
        coord -= origo

        for c in coord:
            c[3], c[4] = -c[4], c[3]

        coord += origo
        return coord.flatten().tolist()

    @classmethod
    def __translate_lower_coord(cls, coord, translation):
        coord = numpy.array([
            list(map(float, coord[i:i+6:]))
            for i in range(0, len(coord), 6)
            ])
        translation = numpy.array(3*[0.] + list(translation))

        coord = coord + translation
        return coord.flatten().tolist()

    @classmethod
    def assert_coord(cls, nx, ny, nz, coord, negative_values=False):
        """

        Raises an AssertionError if the coord is not as expected. In
        particular, it is verified that:

            - coord has the approperiate length (6*(nx+1)*(ny+1)) and
            - that all values are positive unless negative_values are
              explicitly allowed.

        """

        if len(coord) != 6*(nx+1)*(ny+1):
            raise AssertionError(
                    "Expected len(coord) to be %d, was %d" %
                    (6*(nx+1)*(ny+1), len(coord))
                    )

        if not negative_values and min(coord) < 0:
            raise AssertionError("Negative COORD values was generated. " +
                    "This is likely due to a tranformation. " +
                    "Increasing the escape_origio_shift will most likely " +
                    "fix the problem")

    @classmethod
    def assert_actnum(cls, nx, ny, nz, actnum):
        """

        Raises an AssertionError if the actnum is not as expected. In
        particular, it is verified that:

            - actnum has the approperiate length nx*ny*nz and
            - that all values are either 0 or 1.

        """

        if actnum is None:
            return

        if len(actnum) != nx*ny*nz:
            raise AssertionError(
                    "Expected the length of ACTNUM to be %d, was %s."
                    %(nx*ny*nz, len(actnum))
                    )

        if set(actnum)-set([0,1]):
            raise AssertionError(
                "Expected ACTNUM to consist of 0's and 1's, was %s."
                % ", ".join(map(str, set(actnum)))
                )

    @classmethod
    def extract_coord(cls, dims, coord, ijk_bounds):
        nx, ny, nz = dims
        (lx, ux), (ly, uy), (lz, uz) = ijk_bounds
        new_nx, new_ny, new_nz = ux-lx+1, uy-ly+1, uz-lz+1

        cls.assert_coord(nx, ny, nz, coord, negative_values=True)

        # Format COORD
        coord = divide(divide(coord, 6), nx+1)

        # Extract new COORD
        new_coord = [coord_slice[lx:ux+2:]
                        for coord_slice in coord[ly:uy+2]]

        # Flatten and verify
        new_coord = flatten(flatten(new_coord))
        cls.assert_coord(new_nx, new_ny, new_nz, new_coord,
                negative_values=True)

        return construct_floatKW("COORD", new_coord)

    @classmethod
    def extract_zcorn(cls, dims, zcorn, ijk_bounds):
        nx, ny, nz = dims
        (lx, ux), (ly, uy), (lz, uz) = ijk_bounds
        new_nx, new_ny, new_nz = ux-lx+1, uy-ly+1, uz-lz+1

        cls.assert_zcorn(nx, ny, nz, zcorn, twisted_check=False)

        # Format ZCORN
        zcorn = divide(divide(zcorn, 2*nx), 2*ny)

        # Extract new ZCORN
        new_zcorn = [
                        y_slice[2*lx:2*ux+2:]
                        for z_slice in zcorn[2*lz:2*uz+2:]
                        for y_slice in z_slice[2*ly:2*uy+2:]
                    ]

        # Flatten and verify
        new_zcorn = flatten(new_zcorn)
        cls.assert_zcorn(new_nx, new_ny, new_nz, new_zcorn)

        return construct_floatKW("ZCORN", new_zcorn)

    @classmethod
    def extract_actnum(cls, dims, actnum, ijk_bounds):
        if actnum is None:
            return None

        nx, ny, nz = dims
        (lx, ux), (ly, uy), (lz, uz) = ijk_bounds
        new_nx, new_ny, new_nz = ux-lx+1, uy-ly+1, uz-lz+1

        cls.assert_actnum(nx, ny, nz, actnum)

        actnum = divide(divide(actnum, nx), ny)

        new_actnum = [
                        y_slice[lx:ux+1:]
                        for z_slice in actnum[lz:uz+1:]
                        for y_slice in z_slice[ly:uy+1:]
                    ]

        new_actnum = flatten(new_actnum)
        cls.assert_actnum(new_nx, new_ny, new_nz, new_actnum)

        actnumkw = EclKW("ACTNUM", len(new_actnum), EclDataType.ECL_INT)
        for i, value in enumerate(new_actnum):
            actnumkw[i] = value

        return actnumkw

    @classmethod
    def __translate_coord(cls, coord, translation):
        coord = numpy.array([
            list(map(float, coord[i:i+6:]))
            for i in range(0, len(coord), 6)
            ])
        translation = numpy.array(list(translation) + list(translation))

        coord = coord + translation
        return construct_floatKW("COORD", coord.flatten().tolist())


    @classmethod
    def extract_subgrid(cls, grid, ijk_bounds,
            decomposition_change=False, translation=None):

        """
        Extracts a subgrid from the given grid according to the specified
        bounds.

        @ijk_bounds: The bounds describing the subgrid. Should be a tuple of
        length 3, where each element gives the bound for the i, j, k
        coordinates of the subgrid to be described, respectively. Each bound
        should either be an interval of the form (a, b) where 0 <= a <= b < nx
        or a single integer a which is equivialent to the bound (a, a).

        NOTE: The given bounds are including endpoints.

        @decomposition_change: Depending on the given ijk_bounds, libecl might
        decompose the cells of the subgrid differently when extracted from
        grid. This is somewhat unexpected behaviour and if this event occur we
        give an exception together with an description for how to avoid this,
        unless decompostion_change is set to True.

        @translation: Gives the possibility of translating the subgrid. Should
        be given as a tuple (dx, dy, dz), where each coordinate of the grid
        will be moved by di in direction i.

        """

        gdims = grid.getDims()[:-1:]
        nx, ny, nz = gdims
        ijk_bounds = cls.assert_ijk_bounds(gdims, ijk_bounds)

        coord = grid.export_coord()
        cls.assert_coord(nx, ny, nz, coord, negative_values=True)

        zcorn = grid.export_zcorn()
        cls.assert_zcorn(nx, ny, nz, zcorn)

        actnum = grid.export_actnum()
        cls.assert_actnum(nx, ny, nz, actnum)

        mapaxes = grid.export_mapaxes()

        sub_data = cls.extract_subgrid_data(
                                    gdims,
                                    coord, zcorn,
                                    ijk_bounds=ijk_bounds,
                                    actnum=actnum,
                                    mapaxes=mapaxes,
                                    decomposition_change=decomposition_change,
                                    translation=translation
                                    )

        sdim = tuple([b-a+1 for a,b in ijk_bounds])
        sub_coord, sub_zcorn, sub_actnum = sub_data

        return EclGrid.create(sdim, sub_zcorn, sub_coord, sub_actnum, mapaxes=mapaxes)

    @classmethod
    def extract_subgrid_data(cls, dims, coord, zcorn, ijk_bounds, actnum=None,
            mapaxes=None, decomposition_change=False, translation=None):
        """

        Extracts subgrid data from COORD, ZCORN and potentially ACTNUM. It
        returns similar formatted data for the subgrid described by the bounds.

        @dims: The dimentions (nx, ny, nz) of the grid

        @coord: The COORD data of the grid.

        @zcorn: The ZCORN data of the grid.

        @ijk_bounds: The bounds describing the subgrid. Should be a tuple of
        length 3, where each element gives the bound for the i, j, k
        coordinates of the subgrid to be described, respectively. Each bound
        should either be an interval of the form (a, b) where 0 <= a <= b < nx
        or a single integer a which is equivialent to the bound (a, a).

        NOTE: The given bounds are including endpoints.

        @actnum: The ACTNUM data of the grid.

        @mapaxes The MAPAXES data of the grid.

        @decomposition_change: Depending on the given ijk_bounds, libecl might
        decompose the cells of the subgrid differently when extracted from
        grid. This is somewhat unexpected behaviour and if this event occur we
        give an exception together with an description for how to avoid this,
        unless decompostion_change is set to True.

        @translation: Gives the possibility of translating the subgrid. Should
        be given as a tuple (dx, dy, dz), where each coordinate of the grid
        will be moved by di in direction i.

        """
        coord, zcorn = list(coord), list(zcorn)
        actnum = None if actnum is None else list(actnum)

        ijk_bounds = cls.assert_ijk_bounds(dims, ijk_bounds)
        cls.assert_decomposition_change(ijk_bounds, decomposition_change)

        nx, ny, nz = dims
        (lx, ux), (ly, uy), (lz, uz) = ijk_bounds
        new_nx, new_ny, new_nz = ux-lx+1, uy-ly+1, uz-lz+1

        new_coord = cls.extract_coord(dims, coord, ijk_bounds)
        new_zcorn = cls.extract_zcorn(dims, zcorn, ijk_bounds)
        new_actnum = cls.extract_actnum(dims, actnum, ijk_bounds)

        if translation is not None:
            mtranslation = pre_mapaxes_translation(translation, mapaxes)
            new_coord = cls.__translate_coord(new_coord, mtranslation)

            for i in range(len(new_zcorn)):
                new_zcorn[i] += translation[2]

        return new_coord, new_zcorn, new_actnum

    @classmethod
    def assert_ijk_bounds(cls, dims, ijk_bounds):
        ijk_bounds = list(ijk_bounds)

        for i in range(len(ijk_bounds)):
            if isinstance(ijk_bounds[i], int):
                ijk_bounds[i] = [ijk_bounds[i]]
            if len(ijk_bounds[i]) == 1:
                ijk_bounds[i] += ijk_bounds[i]

        if len(ijk_bounds) != 3:
            raise ValueError(
                    "Expected ijk_bounds to contain three intervals, " +
                    "contained only %d" % len(ijk_bounds))

        for n, bound in zip(dims, ijk_bounds):
            if len(bound) != 2:
                raise ValueError(
                    "Expected bound to consist of two elements, was %s",
                    str(bound))

            if not (isinstance(bound[0], int) and isinstance(bound[1], int)):
                raise TypeError(
                    "Expected bound to consist of two integers, ",
                    "was %s (%s)"
                    %(str(bound), str((map(type,bound))))
                    )

            if not (0 <= bound[0] <= bound[1] < n):
                raise ValueError(
                    "Expected bounds to have the following format: " +
                    "0 <= lower bound <= upper_bound < ni, "+
                    "was %d <=? %d <=? %d <? %d."
                    % (0, bound[0], bound[1], n))

        return ijk_bounds

    @classmethod
    def assert_decomposition_change(cls, ijk_bounds, decomposition_change):
        if sum(list(zip(*ijk_bounds))[0]) % 2 == 1 and not decomposition_change:
            raise ValueError(
                    "The subgrid defined by %s " % str(ijk_bounds) +
                    "will cause an unintended decomposition change. " +
                    "Either change one of the lower bounds by 1 " +
                    "or activate decomposition_change."
                    )

monkey_the_camel(EclGridGenerator, 'createRectangular', EclGridGenerator.create_rectangular, classmethod)



#  Copyright (C) 2013  Statoil ASA, Norway.
#
#  The file 'ecl_npv.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

import re
import datetime
import numbers

from ecl.util.util import monkey_the_camel
from ecl.summary import EclSum


class NPVParseKey(object):
    def __init__(self, eclNPV):
        self.baseCase = eclNPV.baseCase
        self.NPV = eclNPV


    def __call__(self, matchObject):
        key = matchObject.group(1)
        smspecNode = self.baseCase.smspec_node(key)
        if smspecNode.isTotal():
            var = key.replace(":", "_")
            self.NPV.addKey(key, var)
            return var + "[i]"
        else:
            raise ValueError("Sorry - the key: %s is not a total key - aborting" % key)


class NPVPriceVector(object):
    def __init__(self, argList):
        self.dateList = []
        if isinstance(argList, list):
            for item in argList:
                if isinstance(item, tuple) and len(item) == 2:
                    self.addItem(item)
                else:
                    raise ValueError("Each element in list must be tuple with two elements")
        else:
            raise ValueError("Constructor argument must be list")



    def add_item(self, item):
        dateItem = item[0]
        try:
            year = dateItem.year
            month = dateItem.month
            day = dateItem.day
            date = datetime.date(year, month, day)
        except AttributeError:
            try:
                tmp = re.split("[ .-/]", dateItem)
                day = int(tmp[0])
                month = int(tmp[1])
                year = int(tmp[2])

                date = datetime.date(year, month, day)
            except Exception:
                raise ValueError("First element was invalid date item")


        if len(self.dateList):
            prevItem = self.dateList[-1]
            if prevItem[0] >= date:
                raise ValueError("The dates must be in a strictly increasing order")

        value = item[1]
        if isinstance(value, numbers.Real) or callable(value):
            self.dateList.append((date, value))
        else:
            raise ValueError("The value argument must be a scalar number or callable")

    @staticmethod
    def assert_date(date):
        try:
            year = date.year
            month = date.month
            day = date.day
            date = datetime.date(year, month, day)

            return date
        except AttributeError:
            return date.date()



    def eval_date(self, dateItem, date):
        value = dateItem[1]
        if callable(value):
            td = date - dateItem[0]
            return value(td.days)
        else:
            return value


    def eval(self, date):
        date = self.assertDate(date)
        startDate = self.dateList[0][0]
        if date >= startDate:
            endDate = self.dateList[-1][0]
            if date >= endDate:
                return self.evalDate(self.dateList[-1], date)
            else:
                index1 = 0
                index2 = len(self.dateList) - 1
                while True:
                    if (index2 - index1) == 1:
                        index = index1
                        break

                    index = (index1 + index2) >> 1
                    item = self.dateList[index]
                    if date >= item[0]:
                        index1 = index
                    else:
                        index2 = index
                return self.evalDate(self.dateList[index], date)
        else:
            raise ValueError("Input date:%s before start of vector" % date)





class EclNPV(object):
    sumKeyRE = re.compile("[[]([\w:,]+)[]]")


    def __init__(self, baseCase):
        sum = EclSum(baseCase)
        if sum:
            self.baseCase = sum
        else:
            raise Error("Failed to open ECLIPSE sumamry case:%s" % baseCase)
        self.expression = None
        self.keyList = {}
        self.start = None
        self.end = None
        self.interval = "1Y"


    def eval(self):
        if self.expression is None:
            raise ValueError("Can not eval with an expression to evaluate")
        pass


    def get_expression(self):
        return self.expression


    def set_expression(self, expression):
        self.compiled_expr = self.compile(expression)
        self.expression = expression


    def get_key_list(self):
        return self.keyList.keys()


    def add_key(self, key, var):
        self.keyList[key] = var


    def parse_expression(self, expression):
        self.keyList = {}
        if expression.count("[") != expression.count("]"):
            raise ValueError("Expression:%s invalid - not matching [ and ]" % expression)

        replaceKey = NPVParseKey(self)
        parsedExpression = self.sumKeyRE.sub(replaceKey, expression)
        return parsedExpression


    def compile(self, expression):
        parsedExpression = self.parseExpression(expression)
        self.code = "trange = self.baseCase.timeRange(self.start, self.end, self.interval)\n"
        for (key,var) in self.keyList.items():
            self.code += "%s = self.baseCase.blockedProduction(\"%s\", trange)\n" % (var, key)

        self.code += "npv = 0\n"
        self.code += """
for i in range(len(trange) - 1):
   npv += %s
varDict[\"npv\"] = npv
""" % parsedExpression



    def eval_npv(self):
        byteCode = compile(self.code, "<string>", 'exec')
        varDict = {}
        eval(byteCode)
        return varDict["npv"]



monkey_the_camel(NPVPriceVector, 'addItem', NPVPriceVector.add_item)
monkey_the_camel(NPVPriceVector, 'assertDate', NPVPriceVector.assert_date, staticmethod)
monkey_the_camel(NPVPriceVector, 'evalDate', NPVPriceVector.eval_date)

monkey_the_camel(EclNPV, 'getExpression', EclNPV.get_expression)
monkey_the_camel(EclNPV, 'setExpression', EclNPV.set_expression)
monkey_the_camel(EclNPV, 'getKeyList', EclNPV.get_key_list)
monkey_the_camel(EclNPV, 'addKey', EclNPV.add_key)
monkey_the_camel(EclNPV, 'parseExpression', EclNPV.parse_expression)
monkey_the_camel(EclNPV, 'evalNPV', EclNPV.eval_npv)



from ecl import EclPrototype


def installAbortSignals():
    install_signals()


def updateAbortSignals():
    """
    Will install the util_abort_signal for all UNMODIFIED signals.
    """
    update_signals()


install_signals = EclPrototype("void util_install_signals()")
update_signals = EclPrototype("void util_update_signals()")



#  Copyright (C) 2014  Statoil ASA, Norway. 
#   
#  The file 'cpolyline_collection.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details. 
"""
Create a polygon
"""
import ctypes

from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.util.geometry import CPolyline


class CPolylineCollection(BaseCClass):
    TYPE_NAME = "geo_polygon_collection"

    _alloc_new          = EclPrototype("void*            geo_polygon_collection_alloc(  )" , bind = False)
    _free               = EclPrototype("void             geo_polygon_collection_free( geo_polygon_collection )" )
    _size               = EclPrototype("int              geo_polygon_collection_size( geo_polygon_collection)" )
    _create_polyline    = EclPrototype("geo_polygon_ref  geo_polygon_collection_create_polygon(geo_polygon_collection , char*)" )
    _has_polyline       = EclPrototype("bool             geo_polygon_collection_has_polygon(geo_polygon_collection , char*)" )
    _iget               = EclPrototype("geo_polygon_ref  geo_polygon_collection_iget_polygon(geo_polygon_collection , int)" )
    _get                = EclPrototype("geo_polygon_ref  geo_polygon_collection_get_polygon(geo_polygon_collection , char*)" )
    _add_polyline       = EclPrototype("void             geo_polygon_collection_add_polygon(geo_polygon_collection , geo_polygon , bool)")



    def __init__(self):
        c_ptr = self._alloc_new(  )
        super(CPolylineCollection , self).__init__( c_ptr )
        self.parent_ref = None


    def __contains__(self , name):
        return self._has_polyline(name)


    def __len__(self):
        return self._size( )


    def __iter__(self):
        index = 0

        while index < len(self):
            yield self[index]
            index += 1


    def __getitem__(self , index):
        if isinstance(index , int):
            if index < 0:
                index += len(self)

            if 0 <= index < len(self):
                return self._iget( index).setParent( self )
            else:
                raise IndexError("Invalid index:%d - valid range: [0,%d)" % (index , len(self)))
        elif isinstance(index , str):
            if index in self:
                return self._get(index)
            else:
                raise KeyError("No polyline named:%s" % index)
        else:
            raise TypeError("The index argument must be string or integer")


    def shallowCopy(self):
        copy = CPolylineCollection()
        for pl in self:
            copy._add_polyline(pl , False)

        # If we make a shallow copy we must ensure that source, owning
        # all the polyline objects does not go out of scope.
        copy.parent_ref = self
        return copy



    def addPolyline(self , polyline , name = None):
        if not isinstance(polyline , CPolyline):
            polyline = CPolyline( init_points = polyline , name = name)
        else:
            if not name is None:
                raise ValueError("The name keyword argument can only be supplied when add not CPOlyline object")

        name = polyline.getName()
        if name and name in self:
            raise KeyError("The polyline collection already has an object:%s" % name)

        if polyline.isReference():
            self._add_polyline( polyline , False)
        else:
            polyline.convertToCReference( self )
            self._add_polyline( polyline , True)



    def createPolyline(self , name = None):
        if name and name in self:
            raise KeyError("The polyline collection already has an object:%s" % name)

        polyline = self._create_polyline(name)
        polyline.setParent( parent = self )
        return polyline


    def free(self):
        self._free( )



import os
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        cmake_args = [] # Fill in extra stuff we may need
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir]
        cfg = 'Debug' if self.debug else 'Release'
        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        build_args = ['--config', cfg, '--', '-j2']
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        env = os.environ.copy()
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)

setup(
    name='dhan16opm_libecl',
    version='0.1.5',
    author_email='chandan.nath@gmail.com',
    description='libecl',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/statoil/libecl",
    license="GNU General Public License, Version 3, 29 June 2007",
    packages=find_packages(where='python', exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_dir={'': 'python'},
    ext_package='ecl',
    ext_modules=[CMakeExtension('libecl')],
    cmdclass=dict(build_ext=CMakeBuild),
    install_requires=[
        'cwrap',
        'numpy',
        'pandas',
    ],
)



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_util.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Constants from the header ecl_util.h - some stateless functions.

This module does not contain any class definitions; it mostly consists
of enum definitions/values from ecl_util.h; the enum values are
extracted from the shared library in a semi-automagic manner using the
BaseCEnum class from cwrap.

In addition to the enum definitions there are a few stateless
functions from ecl_util.c which are not bound to any class type.
"""

import ctypes

from cwrap import BaseCEnum
from ecl.util.util import monkey_the_camel
from ecl import EclPrototype

class EclFileEnum(BaseCEnum):
    TYPE_NAME = "ecl_file_enum"
    ECL_OTHER_FILE = None
    ECL_RESTART_FILE = None
    ECL_UNIFIED_RESTART_FILE = None
    ECL_SUMMARY_FILE = None
    ECL_UNIFIED_SUMMARY_FILE = None
    ECL_GRID_FILE = None
    ECL_EGRID_FILE = None
    ECL_INIT_FILE = None
    ECL_RFT_FILE = None
    ECL_DATA_FILE = None


EclFileEnum.addEnum("ECL_OTHER_FILE",           0)
EclFileEnum.addEnum("ECL_RESTART_FILE",         1)
EclFileEnum.addEnum("ECL_UNIFIED_RESTART_FILE", 2)
EclFileEnum.addEnum("ECL_SUMMARY_FILE",         4)
EclFileEnum.addEnum("ECL_UNIFIED_SUMMARY_FILE", 8)
EclFileEnum.addEnum("ECL_SUMMARY_HEADER_FILE",  16)
EclFileEnum.addEnum("ECL_GRID_FILE",            32)
EclFileEnum.addEnum("ECL_EGRID_FILE",           64)
EclFileEnum.addEnum("ECL_INIT_FILE",            128)
EclFileEnum.addEnum("ECL_RFT_FILE",             256)
EclFileEnum.addEnum("ECL_DATA_FILE",            512)


#-----------------------------------------------------------------

class EclPhaseEnum(BaseCEnum):
    TYPE_NAME="ecl_phase_enum"
    ECL_OIL_PHASE = None
    ECL_GAS_PHASE = None
    ECL_WATER_PHASE = None

EclPhaseEnum.addEnum("ECL_OIL_PHASE", 1)
EclPhaseEnum.addEnum("ECL_GAS_PHASE", 2)
EclPhaseEnum.addEnum("ECL_WATER_PHASE", 4)


#-----------------------------------------------------------------

class EclUnitTypeEnum(BaseCEnum):
    TYPE_NAME = "ecl_unit_enum"

    ECL_METRIC_UNITS = None
    ECL_FIELD_UNITS  = None
    ECL_LAB_UNITS    = None
    ECL_PVT_M_UNITS  = None

EclUnitTypeEnum.addEnum("ECL_METRIC_UNITS", 1)
EclUnitTypeEnum.addEnum("ECL_FIELD_UNITS", 2)
EclUnitTypeEnum.addEnum("ECL_LAB_UNITS", 3)
EclUnitTypeEnum.addEnum("ECL_PVT_M_UNITS", 4)



#-----------------------------------------------------------------

class EclFileFlagEnum(BaseCEnum):
    TYPE_NAME="ecl_file_flag_enum"
    ECL_FILE_CLOSE_STREAM = None
    ECL_FILE_WRITABLE = None

EclFileFlagEnum.addEnum("ECL_FILE_CLOSE_STREAM", 1)
EclFileFlagEnum.addEnum("ECL_FILE_WRITABLE", 2)


#-----------------------------------------------------------------

class EclUtil(object):
    _get_num_cpu    = EclPrototype("int ecl_util_get_num_cpu(char*)", bind = False)
    _get_file_type  = EclPrototype("ecl_file_enum ecl_util_get_file_type(char*, bool*, int*)", bind = False)
    _get_start_date = EclPrototype("time_t ecl_util_get_start_date(char*)", bind = False)
    _get_report_step = EclPrototype("int ecl_util_filename_report_nr(char*)", bind = False)


    @staticmethod
    def get_num_cpu(datafile):
        """
        Parse ECLIPSE datafile and determine how many CPUs are needed.

        Will look for the "PARALLELL" keyword, and then read off the
        number of CPUs required. Will return one if no PARALLELL keyword
        is found.
        """
        return EclUtil._get_num_cpu(datafile)

    @staticmethod
    def get_file_type(filename):
        """
        Will inspect an ECLIPSE filename and return an integer type flag.
        """
        file_type, fmt, step = EclUtil.inspectExtension(filename)
        return file_type

    @staticmethod
    def get_start_date(datafile):
        return EclUtil._get_start_date(datafile).datetime()

    @staticmethod
    def inspect_extension(filename):
        """Will inspect an ECLIPSE filename and return a tuple consisting of
        file type (EclFileEnum), a bool for formatted or not, and an
        integer for the step number.

        """
        fmt_file = ctypes.c_bool()
        report_step = ctypes.c_int(-1)
        file_type = EclUtil._get_file_type(filename, ctypes.byref(fmt_file), ctypes.byref(report_step))
        if report_step.value == -1:
            step = None
        else:
            step = report_step.value

        return (file_type, fmt_file.value, step)

    @staticmethod
    def report_step(filename):
        report_step = EclUtil._get_report_step(filename)
        if report_step < 0:
            raise ValueError("Could not infer report step from: %s" % filename)

        return report_step



get_num_cpu = EclUtil.get_num_cpu
get_file_type = EclUtil.get_file_type
get_start_date = EclUtil.get_start_date

monkey_the_camel(EclUtil, 'inspectExtension', EclUtil.inspect_extension, staticmethod)
monkey_the_camel(EclUtil, 'reportStep', EclUtil.report_step, staticmethod)



#  Copyright (C) 2016  Statoil ASA, Norway. 
#   
#  The file 'temp_area.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details.

import os
import os.path
from   ecl import EclPrototype
from . import TestArea

class TempArea(TestArea):
    """TempArea class is essentially similar to the TestArea class, with
    the only difference that the cwd is *not* changed into the newly
    created area.
    """
    
    _temp_area_alloc           = EclPrototype("void* temp_area_alloc( char* )" , bind = False)
    _temp_area_alloc_relative  = EclPrototype("void* temp_area_alloc_relative( char* , char* )" , bind = False)

    def __init__(self, name, prefix = None , store_area=False):
        if prefix:
            if os.path.exists( prefix ):
                c_ptr = self._temp_area_alloc_relative(prefix , name)
            else:
                raise IOError("The prefix path:%s must exist" % prefix)
        else:
            c_ptr = self._temp_area_alloc(name)
        super(TempArea, self).__init__(name , c_ptr = c_ptr , store_area = store_area)


    def __str__(self):
        return self.getPath()

    
    def get_cwd(self):
        """
        Since the TempArea class does *not* change the cwd this method
        just returns the ordinary os.getcwd().
        """
        return os.getcwd()

    
    def getPath(self):
        """
        Will return the full path to the temporary working area.
        """
        return self._get_cwd( )

    

class TempAreaContext(object):
    def __init__(self, name, prefix = None , store_area=False):
        self.name = name
        self.store_area = store_area
        self.prefix = prefix

    def __enter__(self):
        """
        @rtype: TempArea
        """
        self.temp_area = TempArea(self.name, prefix = self.prefix , store_area = self.store_area )
        return self.temp_area


    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.temp_area
        return False
    

    



import os
import re

class SourceEnumerator(object):

    @classmethod
    def removeComments(cls, code_string):
        code_string = re.sub(re.compile("/\*.*?\*/",re.DOTALL ) ,"" ,code_string) # remove all occurance streamed comments (/*COMMENT */) from string
        code_string = re.sub(re.compile("//.*?\n" ) ,"" ,code_string) # remove all occurance singleline comments (//COMMENT\n ) from string
        return code_string

    
    @classmethod
    def findEnum(cls, enum_name, full_source_file_path):
        with open(full_source_file_path, "r") as f:
            text = f.read()

        text = SourceEnumerator.removeComments(text)

        enum_pattern = re.compile("typedef\s+enum\s+\{(.*?)\}\s*(\w+?);", re.DOTALL)

        for enum in enum_pattern.findall(text):
            if enum[1] == enum_name:
                return enum[0]

        raise ValueError("Enum with name: '%s' not found!" % enum_name)


    @classmethod
    def findEnumerators(cls, enum_name, source_file):
        enum_text = SourceEnumerator.findEnum(enum_name, source_file)

        enumerator_pattern = re.compile("(\w+?)\s*?=\s*?(\d+)")

        enumerators = []
        for enumerator in enumerator_pattern.findall(enum_text):
            enumerators.append((enumerator[0], int(enumerator[1])))

        return enumerators



#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl.eclfile import EclFile, Ecl3DKW


class Ecl3DFile(EclFile):

    def __init__(self, grid, filename, flags=0):
        self.grid = grid
        super(Ecl3DFile, self).__init__(filename, flags)


    def __getitem__(self, index):
        return_arg = super(Ecl3DFile, self).__getitem__(index)
        if isinstance(return_arg,list):
            kw_list = return_arg
        else:
            kw_list = [return_arg]

        # Go through all the keywords and try inplace promotion to Ecl3DKW
        for kw in kw_list:
            try:
                Ecl3DKW.castFromKW(kw, self.grid)
            except ValueError:
                pass

        return return_arg



import os
import shutil

class PathContext(object):
    def __init__(self , path , store = False):
        self.path = path
        self.cwd = os.getcwd()
        self.store = store
        self.path_list = [ ]

        if not os.path.exists(path):
            work_path = path

            while True:
                work_path , base = os.path.split(work_path)
                if work_path:
                    if os.path.isdir(work_path):
                        break
                    else:
                        self.path_list.append( work_path )
                else:
                    break

            os.makedirs( path )
        else:
            if not self.store:
                raise OSError("Entry %s already exists" % path)
        os.chdir( path )



    def __exit__(self , exc_type , exc_val , exc_tb):
        os.chdir( self.cwd )
        if self.store == False:
            shutil.rmtree( self.path )
            for path in self.path_list:
                try:
                    os.rmdir( path )
                except OSError:
                    break

        return False


    def __enter__(self):
        return self



#!/usr/bin/env python

from os import listdir
from os.path import isfile, join, isdir, islink
import sys



def findFilesAndDirectories(directory):
    all_files = listdir(directory)
    files = []
    directories = []
    for f in all_files:
        path = join(directory, f)
        if isfile(path) and not f == "CMakeLists.txt" and not islink(path):
            files.append(f)
        if isdir(path):
            directories.append(f)

    return sorted(files), sorted(directories)


def findRelativeModulePath(directory):
    """@type directory: str"""
    index = directory.rfind("python/")
    index += len("python/")
    return directory[index:len(directory)]

def createPythonSources(files, test_sources=False):
    result = ""

    if len(files) > 0:
        result = "set(%s\n" % ("PYTHON_SOURCES" if not test_sources else "TEST_SOURCES")

    files = [f for f in files if f.endswith(".py")]

    for f in files:
        result += "    " + str(f) + "\n"

    if len(files) > 0:
        result += ")"

    return result

def addSubDirectories(directories):
    result = ""

    for d in directories:
        result += "add_subdirectory(" + str(d) + ")\n"

    return result

def addPythonPackage(relative_module_path, test_sources=False):
    module_name = ".".join(relative_module_path.split("/"))
    source_type = "PYTHON_SOURCES" if not test_sources else "TEST_SOURCES"
    template = "add_python_package(\"python.%s\" ${PYTHON_INSTALL_PREFIX}/%s \"${%s}\" %s)"

    install = "False" if test_sources else "True"

    return template % (module_name, relative_module_path, source_type, install)

def addInclude(filename):
    with open(filename, "r") as include_file:
        content = include_file.read()
    return content

files, directories = findFilesAndDirectories(sys.argv[1])
module_path = findRelativeModulePath(sys.argv[1])

output_file = join(sys.argv[1], "CMakeLists.txt")
test_sources = module_path.startswith("tests")
with open(output_file, "w+") as text_file:
    text_file.write(createPythonSources(files, test_sources=test_sources))
    text_file.write("\n\n")
    text_file.write(addPythonPackage(module_path, test_sources=test_sources))
    text_file.write("\n\n")
    text_file.write(addSubDirectories(directories))

    if "local.cmake" in files:
        text_file.write("\n\n")
        text_file.write(addInclude(join(sys.argv[1], "local.cmake")))







from ecl.grid.faults import Layer
from ecl.grid.faults import FaultCollection
from ecl.grid.faults import Fault
from ecl.grid.faults import FaultLine
from ecl.grid.faults import FaultSegment , SegmentMap
from ecl.grid.faults import FaultBlock , FaultBlockCell
from ecl.grid.faults import FaultBlockLayer



#  Copyright (C) 2018  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

"""
  ecl_sum/EclSum: This will load summary results from an ECLIPSE run;
     both data file(s) and the SMSPEC file. The EclSum object can be
     used as basis for queries on summary vectors.
"""


import ecl.util.util
import ecl.util.geometry

from .ecl_sum_var_type import EclSumVarType
from .ecl_sum_tstep import EclSumTStep
from .ecl_sum import EclSum #, EclSumVector, EclSumNode, EclSMSPECNode
from .ecl_sum_keyword_vector import EclSumKeyWordVector
from .ecl_sum_node import EclSumNode
from .ecl_sum_vector import EclSumVector
from .ecl_npv import EclNPV , NPVPriceVector
from .ecl_cmp import EclCmp




from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.well import WellConnectionDirectionEnum

class WellConnection(BaseCClass):
    TYPE_NAME = "well_connection"

    _i                   = EclPrototype("int    well_conn_get_i(well_connection)")
    _j                   = EclPrototype("int    well_conn_get_j(well_connection)")
    _k                   = EclPrototype("int    well_conn_get_k(well_connection)")
    _segment_id          = EclPrototype("int    well_conn_get_segment_id(well_connection)")
    _is_open             = EclPrototype("bool   well_conn_open(well_connection)")
    _is_msw              = EclPrototype("bool   well_conn_MSW(well_connection)")
    _fracture_connection = EclPrototype("bool   well_conn_fracture_connection(well_connection)")
    _matrix_connection   = EclPrototype("bool   well_conn_matrix_connection(well_connection)")
    _connection_factor   = EclPrototype("double well_conn_get_connection_factor(well_connection)")
    _equal               = EclPrototype("bool   well_conn_equal(well_connection, well_connection)")
    _get_dir             = EclPrototype("void*  well_conn_get_dir(well_connection)")
    _oil_rate            = EclPrototype("double well_conn_get_oil_rate(well_connection)")
    _gas_rate            = EclPrototype("double well_conn_get_gas_rate(well_connection)")
    _water_rate          = EclPrototype("double well_conn_get_water_rate(well_connection)")
    _volume_rate         = EclPrototype("double well_conn_get_volume_rate(well_connection)")

    _oil_rate_si         = EclPrototype("double well_conn_get_oil_rate_si(well_connection)")
    _gas_rate_si         = EclPrototype("double well_conn_get_gas_rate_si(well_connection)")
    _water_rate_si       = EclPrototype("double well_conn_get_water_rate_si(well_connection)")
    _volume_rate_si      = EclPrototype("double well_conn_get_volume_rate_si(well_connection)")

    def __init__(self):
        raise NotImplementedError("Class can not be instantiated directly")


    def isOpen(self):
        """ @rtype: bool """
        return self._is_open()


    def ijk(self):
        """ @rtype: tuple of (int, int, int) """
        i = self._i()
        j = self._j()
        k = self._k()
        return i, j, k

    def direction(self):
        """ @rtype: WellConnectionDirectionEnum """
        return self._get_dir()

    def segmentId(self):
        """ @rtype: int """
        return self._segment_id()

    def isFractureConnection(self):
        """ @rtype: bool """
        return self._fracture_connection()

    def isMatrixConnection(self):
        """ @rtype: bool """
        return self._matrix_connection()

    def connectionFactor(self):
        """ @rtype: float """
        return self._connection_factor()

    def __eq__(self, other):
        return self._equal(other)

    def __hash__(self):
        return id(self)

    def __ne__(self, other):
        return not self == other

    def free(self):
        pass

    def isMultiSegmentWell(self):
        """ @rtype: bool """
        return self._is_msw()

    def __repr__(self):
        ijk = str(self.ijk())
        frac = 'fracture ' if self.isFractureConnection() else ''
        open_ = 'open ' if self.isOpen() else 'shut '
        msw = ' (multi segment)' if self.isMultiSegmentWell() else ''
        dir = WellConnectionDirectionEnum(self.direction())
        addr = self._address()
        return 'WellConnection(%s %s%s%s, rates = (O:%s,G:%s,W:%s), direction = %s) at 0x%x' % (ijk, frac, open_, msw, self.oilRate(), self.gasRate(), self.waterRate(), dir, addr)

    def gasRate(self):
        return self._gas_rate()

    def waterRate(self):
        return self._water_rate()

    def oilRate(self):
        return self._oil_rate()

    def volumeRate(self):
        return self._volume_rate()

    def gasRateSI(self):
        return self._gas_rate_si()

    def waterRateSI(self):
        return self._water_rate_si()

    def oilRateSI(self):
        return self._oil_rate_si()

    def volumeRateSI(self):
        return self._volume_rate_si()



#  Copyright (C) 2018  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

"""
  ecl_grid/EclGrid: This will load an ECLIPSE GRID or EGRID file, and
     can then subsequently be used for queries about the grid.

  ecl_region/EclRegion: Convenience class to support selecting cells
     in a grid based on a wide range of criteria. Can be used as a
     mask in operations on EclKW instances.

  ecl_grid_generator/EclGridGenerator: This can be used to generate various
    grids.
"""

import ecl.util.util
import ecl.util.geometry

from .cell import Cell
from .ecl_grid import EclGrid
from .ecl_region import EclRegion
from .ecl_grid_generator import EclGridGenerator



#  Copyright (C) 2011  Statoil ASA, Norway. 
#   
#  The file 'util_func.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details. 
"""
Module with utility functions from util.c
"""

from ecl import EclPrototype

strcmp_int = EclPrototype("int util_strcmp_int( char* , char* )")
"""
Function to compare strings with embedded integers.

Will use proper numeric comparison when comparing strings with
embedded numbers, i.e. "CASE-9" will follow after "CASE-10" when
sorting:

   >> l = ["CASE-9" , "CASE-10"]
   >> l.sort()
   >> print(l)
      ["CASE-10" , "CASE-9"]
   >> l.sort( strcmp_int )
   >> print(l)
      ["CASE-9" , "CASE-10"]

When the standard strcmp() function is used for comparing strings
the '1' will compare as less than the '9' and the order will be
the reverse. Observe that the '-' is not interpreted as a sign
prefix. The strcmp_int function will interpret '.' as separating
character, wheras the strcmp_float() function will interpret '.'
as a descimal point.

@type: (str, str) -> int
"""

strcmp_float = EclPrototype("int util_strcmp_float( char* , char* )")
"""
Function to compare strings with embedded numbers.

See documentation of strcmp_int() for further details.
@type: (str, str) -> int
"""



from ecl.util.geometry import GeoPointset
from ecl.util.geometry import GeoRegion
from ecl.util.geometry import CPolyline
from ecl.util.geometry import CPolylineCollection
from ecl.util.geometry import Polyline
from ecl.util.geometry import XYZIo
from ecl.util.geometry import GeometryTools
from ecl.util.geometry import Surface
